"""Handle logging in a multiprocessing environment.

https://docs.python.org/3/howto/logging-cookbook.html#logging-to-a-single-file-from-multiple-processes
https://github.com/alphatwirl/mantichora/blob/v0.12.0/mantichora/hubmp.py

"""

from __future__ import annotations

import asyncio
from queue import Queue  # noqa F401

import multiprocessing as mp
from multiprocessing.context import BaseContext

from logging import LogRecord, getLogger, DEBUG
from logging.handlers import QueueHandler
from typing import Callable, Optional

from .func import to_thread

__all__ = ["MultiprocessingLogging"]


class MultiprocessingLogging:
    def __init__(self, context: Optional[BaseContext] = None) -> None:
        context = context or mp.get_context()
        self._q: Queue[LogRecord | None] = context.Queue()
        self._task = asyncio.create_task(_listen(self._q))
        self._init = _ConfigureLogger(self._q)

    @property
    def init(self) -> Callable[[], None]:
        """A (picklable) setup function to be called in other processes"""
        return self._init

    async def close(self) -> None:
        await to_thread(self._q.put, None)
        await self._task

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_value, traceback) -> None:
        del exc_type, exc_value, traceback
        await self.close()


class _ConfigureLogger:
    def __init__(self, queue: Queue[LogRecord]):
        self._queue = queue

    def __call__(self) -> None:
        handler = QueueHandler(self._queue)
        logger = getLogger()
        logger.setLevel(DEBUG)
        logger.addHandler(handler)


async def _listen(queue: Queue[LogRecord | None]) -> None:
    while (record := await to_thread(queue.get)) is not None:
        logger = getLogger(record.name)
        if logger.getEffectiveLevel() <= record.levelno:
            logger.handle(record)



import asyncio
import threading

from nextline.utils import ThreadSafeAsyncioEvent, to_thread


async def test_asyncio():
    obj = ThreadSafeAsyncioEvent()
    received = asyncio.Event()
    cleared = asyncio.Event()

    assert not obj.is_set()

    async def send():
        obj.set()
        assert obj.is_set()
        await received.wait()
        obj.clear()
        assert not obj.is_set()
        cleared.set()

    async def receive():
        await obj.wait()
        received.set()
        await cleared.wait()
        assert not obj.is_set()

    await asyncio.gather(send(), receive())


async def test_thread():
    obj = ThreadSafeAsyncioEvent()
    received = threading.Event()
    cleared = threading.Event()

    assert not obj.is_set()

    def send():
        obj.set()
        assert obj.is_set()
        received.wait()
        obj.clear()
        assert not obj.is_set()
        cleared.set()

    async def receive():
        await obj.wait()
        received.set()
        await to_thread(cleared.wait)
        assert not obj.is_set()

    await asyncio.gather(to_thread(send), receive())



import asyncio

import pytest
from unittest.mock import Mock, call

from nextline.utils import ToLoop, to_thread


@pytest.fixture()
async def to_loop():
    y = ToLoop()
    yield y


def func(*_, **__):
    return "result"


async def test_loop(to_loop):
    func_ = Mock(wraps=func)
    ret = to_loop(func_, 1, a="a")
    assert "result" == ret
    assert call(1, a="a") == func_.call_args


async def test_thread(to_loop):
    def test():
        func_ = Mock(wraps=func)
        ret = to_loop(func_, 1, a="a")
        assert "result" == ret
        assert call(1, a="a") == func_.call_args

    await to_thread(test)


async def test_error_no_loop():
    def func():
        # in a thread without an event loop
        with pytest.raises(RuntimeError):
            _ = ToLoop()

    await to_thread(func)


async def test_error_loop_closed():
    async def create():
        return ToLoop()

    def test():

        # create to_loop() while the event loop is running
        to_loop = asyncio.run(create())

        # the loop is closed

        func_ = Mock(wraps=func)
        with pytest.raises(RuntimeError):
            to_loop(func_, 1, a="a")

        assert not func_.called

    await to_thread(test)



from __future__ import annotations

import sys
import threading

from typing import TYPE_CHECKING, Callable, Optional, Tuple, TypeVar


if TYPE_CHECKING:
    from sys import _TraceFunc as TraceFunc

T = TypeVar("T")


def call_with_trace(
    func: Callable[[], T | None],
    trace: TraceFunc,
    thread: Optional[bool] = True,
) -> Tuple[T | None, BaseException | None]:
    """Set the trace function while running the function

    Notes
    -----
    The trace function will be used in all new threads created during the
    function execution regardless of whether the threads are created by the
    function. If the thread option is false, the trace function will not be use
    in any new threads.

    Parameters
    ----------
    func : callable
        A function to be called without any args. Use functools.partial to
        provide args.
    trace: callable
        A trace function.
    thread: bool, default True
        If False, no new threads will be traced.

    Returns
    -------
    tuple
        A tuple with two elements. The first is the return value of func() or
        None if an exception is raised in func(). The second is the exception
        raised in func() or None if no exception is rased.

    """

    ret = None
    exc = None

    trace_org = sys.gettrace()
    if thread:
        threading.settrace(trace)
    sys.settrace(trace)
    try:
        ret = func()
    except BaseException as e:
        exc = e
    finally:
        sys.settrace(trace_org)
        if thread:
            threading.settrace(trace_org)  # type: ignore

        if exc:

            # How to print the exception in the same way as the interpreter.
            # import traceback
            # traceback.print_exception(type(exc), exc, exc.__traceback__)

            if exc.__traceback__:
                # remove this frame from the traceback.
                # Note: exc.__traceback__ is sys._getframe()
                exc.__traceback__ = exc.__traceback__.tb_next

        return ret, exc



from __future__ import annotations

import pytest
from unittest.mock import Mock, call

from nextline.state import State, Running, Finished

from .base import BaseTestState


class TestRunning(BaseTestState):

    state_class = Running

    @pytest.fixture()
    def state(self, running: State) -> State:
        return running

    def test_state(self, state: State, context: Mock):
        super().test_state(state, context)
        assert [call(state)] == context.run.call_args_list

    async def test_finish(self, state):
        finished = await state.finish()
        assert isinstance(finished, Finished)
        await self.assert_obsolete(state)



from __future__ import annotations

from dataclasses import dataclass, field
from operator import itemgetter, attrgetter
from itertools import groupby
from keyword import iskeyword

from unittest.mock import Mock

from typing import (
    Iterable,
    Iterator,
    Mapping,
    Optional,
    Any,
    List,
    Tuple,
    Set,
    TypedDict,
)
from types import FrameType


@dataclass
class TracedScope:
    module: List[str] = field(default_factory=list)
    func: List[str] = field(default_factory=list)


@dataclass
class TraceSummary:
    call: TracedScope = field(default_factory=TracedScope)
    line: TracedScope = field(default_factory=TracedScope)
    return_: TracedScope = field(default_factory=TracedScope)
    exception: TracedScope = field(default_factory=TracedScope)
    opcode: TracedScope = field(default_factory=TracedScope)


class _S(TypedDict):
    event: str
    module: str
    func: str


def summarize_trace_calls(
    mock_trace: Mock,
    modules: Optional[Set[str]] = None,
) -> TraceSummary:
    """Traced modules and functions for each event"""

    args_ = trace_call_args(mock_trace)
    # ((frame, event, arg), ...)

    args = (
        _S(
            event=event,
            module=frame.f_globals.get("__name__", ""),
            func=frame.f_code.co_name,
        )
        for frame, event, _ in args_
    )
    # [{"event": event, "module": module, "func": func}, ...]

    if modules is not None:
        args = (d for d in args if d["module"] in modules)

    args_sorted = sorted(args, key=itemgetter("event"))
    # sort for groupby()

    args_by_events = [
        (
            event + ("_" if iskeyword(event) else ""),  # "return_"
            list(args),  # expand to avoid exhaustion below
        )
        for event, args in groupby(args_sorted, itemgetter("event"))
    ]

    ret = TraceSummary(
        **{
            event: TracedScope(
                **{
                    scope: list(ordered_uniq_values(args, scope))  # type: ignore
                    for scope in ("module", "func")
                }
            )
            for event, args in args_by_events
        }
    )
    return ret


def ordered_uniq_values(
    dicts: Iterable[Mapping[str, str]], key: str
) -> Iterator[str]:
    """List of unique values of dicts for the key

    Note: The order is preserved while consecutive duplicates are removed.
    """
    return (v for v, _ in groupby([d[key] for d in dicts]))


def trace_call_args(trace: Mock) -> Iterator[Tuple[FrameType, str, Any]]:
    return map(attrgetter("args"), trace.call_args_list)



from __future__ import annotations

import pytest

from nextline.state import State, Created, Initialized, Closed

from .base import BaseTestState


class TestCreated(BaseTestState):

    state_class = Created

    @pytest.fixture()
    def state(self, created: State) -> State:
        return created

    async def test_initialize(self, state: State):
        initialized = await state.initialize()
        assert isinstance(initialized, Initialized)
        await self.assert_obsolete(state)

    async def test_close(self, state: State):
        closed = await state.close()
        assert isinstance(closed, Closed)
        await self.assert_obsolete(state)



import pytest
from unittest.mock import Mock

from typing import Set

from nextline.process.trace import TraceSelectFirstModule

from .funcs import TraceSummary

from . import module_a


def test_one(
    target: TraceSummary,
    probe: TraceSummary,
    ref: TraceSummary,
    modules_to_trace: Set[str],
):
    assert ref.call.module
    assert ref.return_.module
    assert ref.call.module == target.call.module
    assert not target.return_.module
    expected = (
        max(
            [
                ref.call.module[ref.call.module.index(m) :]
                for m in modules_to_trace
            ],
            key=len,
        )
        if modules_to_trace
        else []
    )
    assert expected == probe.call.module
    assert set(expected) == set(probe.return_.module)


def f():
    module_a.func_a()


@pytest.fixture()
def func():
    yield f


@pytest.fixture(
    params=[
        set(),
        {__name__},
        {module_a.__name__},
        {__name__, module_a.__name__},
    ]
)
def modules_to_trace(request):
    y = request.param
    yield y


@pytest.fixture()
def target_trace_func(probe_trace_func: Mock, modules_to_trace: Set[str]):
    y = TraceSelectFirstModule(
        trace=probe_trace_func,
        modules_to_trace=modules_to_trace,
    )
    yield y


@pytest.fixture()
def thread():
    yield False



from __future__ import annotations

import asyncio
import dataclasses
from functools import partial
from operator import attrgetter
from itertools import groupby
from collections import Counter
from pathlib import Path
import datetime
from collections import deque

from typing import Optional, Sequence, Set

import pytest

from typing import Dict, Any

from nextline import Nextline
from nextline.utils import agen_with_wait
from nextline.types import RunNo, RunInfo

from .funcs import replace_with_bool


async def test_run(nextline: Nextline, statement: str):
    assert nextline.state == "initialized"

    await asyncio.gather(
        assert_subscriptions(nextline, statement),
        control_execution(nextline),
        run(nextline),
    )

    assert nextline.state == "closed"


async def assert_subscriptions(nextline: Nextline, statement: str):
    await asyncio.gather(
        assert_subscribe_state(nextline),
        assert_subscribe_run_no(nextline),
        assert_subscribe_run_info(nextline, statement),
        assert_subscribe_trace_info(nextline),
        assert_subscribe_prompt_info(nextline),
        # assert_subscribe_stdout(nextline),
    )


async def assert_subscribe_state(nextline: Nextline):
    expected = [
        "initialized",
        "running",
        "finished",
        "initialized",
        "running",
        "finished",
        "closed",
    ]
    actual = [s async for s in nextline.subscribe_state()]
    assert actual == expected


async def assert_subscribe_run_no(nextline: Nextline):
    expected = [1, 2]
    actual = [s async for s in nextline.subscribe_run_no()]
    assert actual == expected


async def assert_subscribe_run_info(nextline: Nextline, statement: str):

    replace: partial[RunInfo] = partial(
        replace_with_bool, fields=("started_at", "ended_at")
    )

    expected_list = deque(
        [
            info := RunInfo(
                run_no=RunNo(1),
                state="running",
                script=statement,
                result=None,
                exception=None,
                started_at=datetime.datetime.now(),
            ),
            dataclasses.replace(
                info,
                state="finished",
                result="null",
                ended_at=datetime.datetime.now(),
            ),
            info := RunInfo(
                run_no=RunNo(2),
                state="running",
                script=statement,
                result=None,
                exception=None,
                started_at=datetime.datetime.now(),
            ),
            dataclasses.replace(
                info,
                state="finished",
                result="null",
                ended_at=datetime.datetime.now(),
            ),
        ]
    )

    async for info in nextline.subscribe_run_info():
        expected = expected_list.popleft()
        expected = replace(expected)
        assert expected == replace(info)
    assert not expected_list


async def assert_subscribe_trace_info(nextline: Nextline):
    results = [s async for s in nextline.subscribe_trace_info()]
    assert {1, 2} == {r.run_no for r in results}
    assert {"running": 10, "finished": 10} == Counter(r.state for r in results)

    groupby_state_ = groupby(
        sorted(results, key=attrgetter("state")),
        attrgetter("state"),
    )
    groupby_state = {k: list(v) for k, v in groupby_state_}
    running = groupby_state["running"]
    finished = groupby_state["finished"]

    assert {1, 2, 3, 4, 5} == {r.trace_no for r in running}
    assert {1, 2, 3, 4, 5} == {r.trace_no for r in finished}

    expected = {(1, None), (2, None), (3, None), (1, 1), (1, 2)}
    assert expected == {(r.thread_no, r.task_no) for r in running}
    assert expected == {(r.thread_no, r.task_no) for r in finished}

    assert all([r.started_at for r in running])
    assert not any([r.ended_at for r in running])

    assert all([r.started_at for r in finished])
    assert all([r.ended_at for r in finished])


async def assert_subscribe_prompt_info(nextline: Nextline):
    results = [s async for s in nextline.subscribe_prompt_info()]
    assert {1, 2} == {r.run_no for r in results}

    assert 382 == len(results)

    expected: Dict[Any, int]
    actual: Counter[Any]

    expected = {1: 144, 2: 78, 3: 78, 4: 50, 5: 32}
    actual = Counter(r.trace_no for r in results)
    assert actual == expected

    assert [r.prompt_no for r in results]

    expected = {True: 116, False: 266}
    actual = Counter(r.open for r in results)
    assert actual == expected

    expected = {"line": 306, "return": 44, "call": 22, "exception": 10}
    actual = Counter(r.event for r in results)
    assert actual == expected

    expected = {True: 382}
    actual = Counter(r.file_name is not None for r in results)
    assert actual == expected

    expected = {True: 382}
    actual = Counter(r.line_no is not None for r in results)
    assert actual == expected

    expected = {True: 232, False: 150}
    actual = Counter(r.stdout is not None for r in results)
    assert actual == expected

    expected = {None: 266, "next": 112, "step": 4}
    actual = Counter(r.command for r in results)
    assert actual == expected

    expected = {True: 232, False: 150}
    actual = Counter(r.started_at is not None for r in results)
    assert actual == expected

    expected = {True: 116, False: 266}
    actual = Counter(r.ended_at is not None for r in results)
    assert actual == expected


async def assert_subscribe_stdout(nextline: Nextline):
    # doesn't work without the `-s` option
    run_no = 1
    results = [s async for s in nextline.subscribe_stdout()]
    assert {run_no} == {r.run_no for r in results}

    assert 1 == len(results)

    r0 = results[0]
    assert 1 == r0.trace_no
    assert "here!\n" == r0.text
    assert r0.written_at


async def run(nextline: Nextline):
    await asyncio.sleep(0.01)
    await nextline.run()
    nextline.exception()
    nextline.result()
    await nextline.reset()
    await nextline.run()
    nextline.exception()
    nextline.result()
    await nextline.close()


async def control_execution(nextline: Nextline):
    prev_ids: Set[int] = set()
    agen = agen_with_wait(nextline.subscribe_trace_ids())
    pending: Sequence[asyncio.Future] = ()
    async for ids_ in agen:
        ids = set(ids_)
        new_ids, prev_ids = ids - prev_ids, ids  # type: ignore

        tasks = {
            asyncio.create_task(control_trace(nextline, id_))
            for id_ in new_ids
        }
        _, pending = await agen.asend(tasks)  # type: ignore

    await asyncio.gather(*pending)


async def control_trace(nextline: Nextline, trace_no):
    # print(f"control_trace({trace_no})")
    file_name = ""
    async for s in nextline.subscribe_prompt_info_for(trace_no):
        # await asyncio.sleep(0.01)
        if not s.open:
            continue
        # print(s)
        if not file_name == s.file_name:
            assert s.file_name
            file_name = s.file_name
            assert nextline.get_source(file_name)
        command = "next"
        if s.event == "line":
            line = nextline.get_source_line(
                line_no=s.line_no,
                file_name=s.file_name,
            )
            command = find_command(line) or command
        nextline.send_pdb_command(command, s.prompt_no, trace_no)
        # await asyncio.sleep(0.01)


def find_command(line: str) -> Optional[str]:
    """The Pdb command indicated in a comment

    For example, returns "step" for the line "func()  # step"
    """
    import re

    if not (comment := extract_comment(line)):
        return None
    regex = re.compile(r"^# +(\w+) *$")
    match = regex.search(comment)
    if match:
        return match.group(1)
    return None


def extract_comment(line: str) -> Optional[str]:
    import io
    import tokenize

    comments = [
        val
        for type, val, *_ in tokenize.generate_tokens(
            io.StringIO(line).readline
        )
        if type == tokenize.COMMENT
    ]
    if comments:
        return comments[0]
    return None


@pytest.fixture
async def nextline(statement):
    async with Nextline(statement) as y:
        yield y


@pytest.fixture
def statement(script_dir, monkey_patch_syspath):
    del monkey_patch_syspath
    return Path(script_dir).joinpath("script.py").read_text()


@pytest.fixture
def monkey_patch_syspath(monkeypatch, script_dir):
    monkeypatch.syspath_prepend(script_dir)
    yield


@pytest.fixture
def script_dir():
    return str(Path(__file__).resolve().parent)



from __future__ import annotations

from concurrent.futures import ThreadPoolExecutor
import queue

import pytest

from nextline.process.run import run, RunArg, set_queues
from nextline.types import RunNo


def test_one(
    expected_exception,
    run_arg: RunArg,
    call_set_queues,
    task_send_commands,
):
    del call_set_queues, task_send_commands
    result, exception = run(run_arg)
    assert result is None
    if expected_exception:
        assert exception
        with pytest.raises(expected_exception):
            raise exception
    else:
        exception is None


@pytest.fixture
def call_set_queues(q_registrar, q_commands):
    set_queues(q_commands, q_registrar)
    yield


@pytest.fixture
def task_send_commands(q_registrar, q_commands):
    with ThreadPoolExecutor(max_workers=1) as executor:
        fut = executor.submit(respond_prompt, q_registrar, q_commands)
        yield
        q_registrar.put(None)
        fut.result()


def respond_prompt(q_registrar, q_commands):
    while (m := q_registrar.get()) is not None:
        key, value, _ = m
        if key != "prompt_info":
            continue
        prompt_info = value
        if prompt_info is None:
            continue
        if not prompt_info.open:
            continue
        q_commands.put(("next", prompt_info.prompt_no, prompt_info.trace_no))


@pytest.fixture
def run_arg(statement: str) -> RunArg:
    y = RunArg(run_no=RunNo(1), statement=statement, filename="<string>")
    return y


@pytest.fixture
def q_registrar():
    return queue.Queue()


@pytest.fixture
def statement(statement_params):
    return statement_params[0]


@pytest.fixture
def expected_exception(statement_params):
    return statement_params[1]


SOURCE_ONE = """
import time
time.sleep(0.001)
""".strip()

SOURCE_COMPILE_ERROR = """
def
""".strip()

SOURCE_RUNTIME_ERROR = """
a
""".strip()

CODE_OBJECT = compile(SOURCE_ONE, "<string>", "exec")


@pytest.fixture(
    params=[
        (SOURCE_ONE, None),
        (SOURCE_COMPILE_ERROR, SyntaxError),
        (SOURCE_RUNTIME_ERROR, NameError),
        (CODE_OBJECT, None),
    ]
)
def statement_params(request):
    return request.param


@pytest.fixture
def q_commands():
    y = queue.Queue()
    return y



import pytest

from nextline.process import script


SOURCE = """
x = 0
""".strip()


def test_code():
    code = compile(SOURCE, "<string>", "exec")
    func = script.compose(code)
    func()


def test_str():
    func = script.compose(SOURCE)
    func()


SOURCE_MODULE_NAME = """
print(__name__)
""".strip()


def test_module_name(capsys):
    # Use stdout to receive data from code
    code = compile(SOURCE_MODULE_NAME, "<string>", "exec")
    func = script.compose(code)
    func()
    actual = capsys.readouterr().out.strip()
    # with capsys.disabled():
    #     print(actual)
    expected = script.__name__
    assert expected == actual


SOURCE_IMPORT = """
import time
def f():
    time.sleep(0.01)
f()
"""


def test_import():
    """Assert that the issue is resolved
    https://github.com/simonsobs/nextline/issues/7
    """
    code = compile(SOURCE_IMPORT, "<string>", "exec")
    func = script.compose(code)
    func()


def test_import_error_demonstration():
    """Show the issue
    https://github.com/simonsobs/nextline/issues/7

    The NameError occurs if the second arg (globals) is not given to
    exec(). An empty dict resolves the issue. Read the doc for the
    detail: https://docs.python.org/3/library/functions.html#exec
    """
    code = compile(SOURCE_IMPORT, "<string>", "exec")
    with pytest.raises(NameError):
        exec(code)
    exec(code, {})



import asyncio

from .loop import ToLoop


class ThreadSafeAsyncioEvent(asyncio.Event):
    """A thread-safe asyncio event

    The methods set() and clear() can be called from any threads. The method
    wait() needs to be called in the thread in which this class is
    instantiated.

    Code originally copied from https://stackoverflow.com/a/33006667/7309855

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._to_loop = ToLoop()

    def set(self):
        return self._to_loop(super().set)

    def clear(self):
        return self._to_loop(super().clear)



__all__ = ["Nextline"]

from .main import Nextline

from ._version import get_versions

__version__ = get_versions()["version"]
"""str: version

The version string, e.g., "0.1.2", "0.1.2+83.ga093a20.dirty".
generated from git tags by versioneer.

Versioneer: https://github.com/warner/python-versioneer

"""

del get_versions



import asyncio
from typing import AsyncIterator, Iterable, TypeVar

T = TypeVar("T")


async def aiterable(iterable: Iterable[T]) -> AsyncIterator[T]:
    '''Wrap iterable so can be used with "async for"'''
    for i in iterable:
        await asyncio.sleep(0)  # let other tasks run
        yield i



from __future__ import annotations
import asyncio
from string import ascii_uppercase

import pytest

from nextline.utils import PubSubItem

from ..aiterable import aiterable


@pytest.fixture()
async def obj():
    async with PubSubItem() as y:
        yield y


def test_close(obj):
    assert obj
    # will be closed in the fixture


async def test_close_multiple_times(obj):
    await obj.close()
    await obj.close()


async def test_get(obj: PubSubItem[int]):
    with pytest.raises(LookupError):
        obj.latest()
    for item in range(5):
        await obj.publish(item)
        assert item == obj.latest()


@pytest.mark.parametrize("n_items", (0, 1, 2, 5))
@pytest.mark.parametrize("n_subscriptions", (0, 1, 2, 5))
async def test_subscribe(obj: PubSubItem[int], n_items, n_subscriptions):
    items = tuple(range(n_items))
    expected = [items] * n_subscriptions

    async def receive():
        return tuple([i async for i in obj.subscribe()])

    async def send():
        async for i in aiterable(items):
            await obj.publish(i)
        await obj.close()

    *results, _ = await asyncio.gather(
        *(receive() for _ in range(n_subscriptions)),
        send(),
    )
    assert results == expected
    assert obj.nsubscriptions == 0


@pytest.mark.parametrize("n", (0, 1, 2, 5))
async def test_nsubscriptions(obj: PubSubItem[int], n):
    async def receive():
        return tuple([i async for i in obj.subscribe()])

    task = asyncio.gather(*(receive() for _ in range(n)))
    await asyncio.sleep(0)
    assert obj.nsubscriptions == n
    await obj.close()
    await task
    assert obj.nsubscriptions == 0


@pytest.mark.parametrize("n_pre_items", (0, 1, 2, 5))
@pytest.mark.parametrize("n_items", (0, 1, 3))
@pytest.mark.parametrize("last", [True, False])
@pytest.mark.parametrize("n_subscriptions", (0, 1, 3))
async def test_last(
    obj: PubSubItem[int | str],
    n_pre_items: int,
    n_items: int,
    last: bool,
    n_subscriptions: int,
):
    pre_items = tuple(ascii_uppercase[:n_pre_items])
    items = tuple(range(n_items))

    expected = [pre_items[-1:] + items if last else items] * n_subscriptions

    async def receive():
        return tuple([i async for i in obj.subscribe(last=last)])

    async def send():
        async for i in aiterable(items):
            await obj.publish(i)
        await obj.close()

    for i in pre_items:
        await obj.publish(i)

    await asyncio.sleep(0.001)

    *results, _ = await asyncio.gather(
        *(receive() for _ in range(n_subscriptions)),
        send(),
    )
    assert results == expected


@pytest.mark.parametrize("n_items", (0, 1, 3))
async def test_put_after_close(obj: PubSubItem[int | str], n_items: int):
    items = tuple(range(n_items))
    for i in items:
        await obj.publish(i)
    await obj.close()
    with pytest.raises(RuntimeError):
        await obj.publish("A")
    if items:
        assert obj.latest() == items[-1]  # the last item is not replaced
    else:
        with pytest.raises(LookupError):
            obj.latest()


@pytest.mark.parametrize("n_items", (0, 1, 3))
async def test_subscribe_after_close(obj: PubSubItem[int], n_items: int):
    items = tuple(range(n_items))
    for i in items:
        await obj.publish(i)
    await obj.close()
    await asyncio.sleep(0.001)

    async def receive():
        return tuple([i async for i in obj.subscribe()])

    results = await receive()
    assert results == ()  # empty


@pytest.mark.parametrize("at", (0, 2, 4))
async def test_break(obj: PubSubItem[int], at: int):
    items = tuple(range(5))
    expected = items[: items.index(at)]

    async def receive():
        ret = []
        assert obj.nsubscriptions == 0
        async for i in obj.subscribe():
            assert obj.nsubscriptions == 1
            if i == at:
                break
            ret.append(i)
        await asyncio.sleep(0.001)
        assert obj.nsubscriptions == 0  # the queue is closed
        return tuple(ret)

    async def send():
        async for i in aiterable(items):
            await obj.publish(i)
        await obj.close()

    results, _ = await asyncio.gather(receive(), send())
    assert results == expected


@pytest.mark.parametrize("n_subscriptions", [0, 1, 2, 5, 50])
@pytest.mark.parametrize("n_pre_items", [0, 1, 2, 50])
@pytest.mark.parametrize("n_post_items", [0, 1, 2, 100])
async def test_matrix(
    obj: PubSubItem[int],
    n_subscriptions: int,
    n_pre_items: int,
    n_post_items: int,
):
    """test if the issue is resolved
    https://github.com/simonsobs/nextline/issues/2

    """

    n_items = n_pre_items + n_post_items

    if n_subscriptions >= 50 and n_items >= 100:
        pytest.skip("nsubscribers >= 50 and nitems >= 100")

    pre_items = tuple(range(-n_pre_items, 0))
    post_items = tuple(range(n_post_items))
    items = pre_items + post_items
    assert len(items) == n_items

    event = asyncio.Event()

    async def receive():
        await event.wait()
        return tuple([i async for i in obj.subscribe()])

    async def send():
        for i in pre_items:
            await obj.publish(i)
        event.set()
        for i in post_items:
            await obj.publish(i)
        await obj.close()

    *results, _ = await asyncio.gather(
        *(receive() for _ in range(n_subscriptions)),
        send(),
    )
    for actual in results:
        if not actual:  # can be empty
            continue
        # print(actual[0])
        expected = items[items.index(actual[0]) :]  # no missing or duplicate
        assert actual == expected



from __future__ import annotations

import asyncio
import dataclasses
from functools import partial
from collections import deque
import datetime

import pytest

from nextline import Nextline
from nextline.types import RunNo, RunInfo

from .funcs import replace_with_bool

STATEMENT = """
import time

time.sleep(100)
""".lstrip()


async def test_run(nextline: Nextline, statement: str):
    assert nextline.state == "initialized"

    await asyncio.gather(
        assert_subscriptions(nextline, statement),
        control(nextline),
        run(nextline),
    )
    assert nextline.state == "closed"


async def assert_subscriptions(nextline: Nextline, statement: str):
    await asyncio.gather(
        assert_subscribe_run_info(nextline, statement),
    )


async def assert_subscribe_run_info(nextline: Nextline, statement: str):

    replace: partial[RunInfo] = partial(
        replace_with_bool, fields=("exception", "started_at", "ended_at")
    )

    expected_list = deque(
        [
            info := RunInfo(
                run_no=RunNo(1),
                state="running",
                script=statement,
                result=None,
                exception=None,
                started_at=datetime.datetime.now(),
            ),
            dataclasses.replace(
                info,
                state="finished",
                exception="KeyboardInterrupt",
                ended_at=datetime.datetime.now(),
            ),
        ]
    )

    async for info in nextline.subscribe_run_info():
        expected = expected_list.popleft()
        if expected.exception:
            assert info.exception
            assert expected.exception in info.exception
        assert replace(expected) == replace(info)
    assert not expected_list


async def run(nextline: Nextline):
    await asyncio.sleep(0.01)
    await nextline.run()
    exc = nextline.exception()
    assert exc
    assert exc.__traceback__  # tblib
    with pytest.raises(KeyboardInterrupt):
        nextline.result()
    await nextline.close()


async def control(nextline: Nextline):
    async for prompt_info in nextline.subscribe_prompt_info():
        if not prompt_info.open:
            continue
        nextline.send_pdb_command(
            "next", prompt_info.prompt_no, prompt_info.trace_no
        )
        if prompt_info.event == "line" and prompt_info.line_no == 3:  # sleep()
            await asyncio.sleep(0.005)
            nextline.interrupt()


@pytest.fixture
async def nextline(statement):
    async with Nextline(statement) as y:
        yield y


@pytest.fixture
def statement():
    return STATEMENT



from __future__ import annotations

import asyncio
from asyncio import Task
from threading import Thread

import pytest
from unittest.mock import Mock

from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Mapping,
    MutableMapping,
    Set,
)

from nextline.process.trace import TraceDispatchThreadOrTask
from nextline.utils import current_task_or_thread
from nextline.utils import ExcThread

if TYPE_CHECKING:
    from sys import _TraceFunc as TraceFunc

from .funcs import TraceSummary, summarize_trace_calls

from . import module_a, module_b


def test_one(
    target: TraceSummary,
    probe: TraceSummary,
    ref: TraceSummary,
    factory: Mock,
    probes: Mapping[Task | Thread, TraceSummary],
    probe_trace_funcs: Mapping[Task | Thread, Mock],
    task_or_threads: Mapping[Task | Thread, Set[Task | Thread]],
    modules_in_summary: Set[str],
):
    assert modules_in_summary is not None
    assert ref.call.module
    assert ref.return_.module
    assert ref.call.func == target.call.func
    assert not target.return_.func

    assert factory.call_count == len(task_or_threads)
    assert probe_trace_funcs.keys() == task_or_threads.keys()
    assert probes.keys() == task_or_threads.keys()

    for created_in, called_in in task_or_threads.items():
        assert {created_in} == called_in

    assert ref.call.func == probe.call.func
    assert ref.return_.func == probe.return_.func


@pytest.fixture()
def modules_in_summary():
    return {__name__, module_a.__name__, module_b.__name__}


def f1():
    module_a.func_a()


def f2():
    t = ExcThread(target=module_a.func_a)
    t.start()
    t.join()


def f3():
    async def a():
        module_a.func_a()

    asyncio.run(a())


def f4():
    t = ExcThread(target=f3)
    t.start()
    t.join()


@pytest.fixture(params=[f1, f2, f3, f4])
def func(request) -> Callable[[], Any]:
    return request.param


@pytest.fixture()
def target_trace_func(factory: Mock) -> TraceFunc:
    return TraceDispatchThreadOrTask(factory=factory)


@pytest.fixture(params=[True, False])
def thread(request) -> bool:
    return request.param


@pytest.fixture()
def probes(
    probe_trace_funcs: MutableMapping[Task | Thread, Mock],
    modules_in_summary: Set[str] | None,
):
    y = {
        k: summarize_trace_calls(v, modules=modules_in_summary)
        for k, v in probe_trace_funcs.items()
    }
    return y


@pytest.fixture()
def factory(
    probe_trace_funcs: MutableMapping[Task | Thread, Mock],
    task_or_threads: MutableMapping[Task | Thread, Set[Task | Thread]],
    probe_trace_func: Mock,
) -> Mock:
    def factory_():
        called_in = set()

        wrap_trace = Mock(wraps=probe_trace_func)

        def trace(*a, **k):
            called_in.add(current_task_or_thread())
            local_trace = probe_trace_func(*a, **k)
            if local_trace is probe_trace_func:
                return wrap_trace
            return local_trace

        wrap_trace.side_effect = trace

        created_in = current_task_or_thread()
        task_or_threads[created_in] = called_in
        probe_trace_funcs[created_in] = wrap_trace
        return wrap_trace

    return Mock(wraps=factory_)


@pytest.fixture()
def probe_trace_funcs() -> Dict[Task | Thread, Mock]:
    return {}


@pytest.fixture()
def task_or_threads():
    return {}



from __future__ import annotations

import pytest
from unittest.mock import Mock

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from sys import _TraceFunc as TraceFunc


def test_wrap_target_trace_func(
    target_trace_func: Mock | TraceFunc,
    wrap_target_trace_func: Mock | TraceFunc,
):
    arg = (Mock(), "", None)
    if target_trace_func is target_trace_func(*arg):
        assert wrap_target_trace_func is wrap_target_trace_func(*arg)
    else:
        assert target_trace_func(*arg) is wrap_target_trace_func(*arg)


@pytest.fixture(params=["self", "another", "none"])
def target_trace_func(request, another_trace_func):
    y = Mock()
    map = {"self": y, "another": another_trace_func, "none": None}
    y.return_value = map[request.param]
    yield y


@pytest.fixture()
def another_trace_func():
    y = Mock()
    y.return_value = y
    yield y



from __future__ import annotations

from tblib import pickling_support  # type: ignore
from typing import TYPE_CHECKING, Optional, Any

if TYPE_CHECKING:
    from .context import Context

pickling_support.install()


class Machine:
    """State machine

                 .-------------.
                 |   Created   |---.
                 '-------------'   |
                       |           |
                       V           |
                 .-------------.   |
            .--->| Initialized |---.
    reset() |    '-------------'   |
            |      |   | run()     |
            |------'   |           |
            |          v           |
            |    .-------------.   |
            |    |   Running   |   |
            |    '-------------'   |
            |          | finish()  |
            |          |           |
            |          V           |
            |    .-------------.   |  close()  .-------------.
            '----|  Finished   |-------------->|   Closed    |
                 '-------------'               '-------------'

    """

    def __init__(self, context: Context):
        self._context = context
        self._state: State = Created(self._context)

    def __repr__(self):
        # e.g., "<Machine 'running'>"
        return f"<{self.__class__.__name__} {self.state_name!r}>"

    @property
    def state_name(self) -> str:
        """e.g., "initialized", "running","""
        try:
            return self._state.name
        except BaseException:
            return "unknown"

    async def initialize(self) -> None:
        """Enter the initialized state"""
        self._state = await self._state.initialize()

    async def run(self) -> None:
        """Enter the running state"""
        self._state = await self._state.run()

    def interrupt(self) -> None:
        self._state.interrupt()

    def terminate(self) -> None:
        self._state.terminate()

    def kill(self) -> None:
        self._state.kill()

    async def finish(self) -> None:
        """Enter the finished state"""
        self._state = await self._state.finish()

    def exception(self) -> Optional[BaseException]:
        ret = self._state.exception()
        return ret

    def result(self) -> Any:
        return self._state.result()

    async def reset(self, *args, **kwargs) -> None:
        """Enter the initialized state"""
        self._state = await self._state.reset(*args, **kwargs)

    async def close(self) -> None:
        """Enter the closed state"""
        self._state = await self._state.close()

    async def __aenter__(self):
        await self.initialize()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        del exc_type, exc_value, traceback
        await self.close()


class StateObsoleteError(Exception):
    """Operation on an obsolete state object."""

    pass


class StateMethodError(Exception):
    """Irrelevant operation on the state."""

    pass


class ObsoleteMixin:
    def assert_not_obsolete(self):
        if self.is_obsolete():
            raise StateObsoleteError(f"The state object is obsolete: {self!r}")

    def is_obsolete(self):
        return getattr(self, "_obsolete", False)

    def obsolete(self):
        self._obsolete = True


class State(ObsoleteMixin):
    """The base state class in the Nextline state machine"""

    name = "state"

    def __init__(self):
        self._context: Context

    def __repr__(self):
        # e.g., "<Initialized 'initialized'>"
        items = [self.__class__.__name__, repr(self.name)]
        if self.is_obsolete():
            items.append("obsolete")
        return f'<{" ".join(items)}>'

    async def initialize(self) -> State:
        self.assert_not_obsolete()
        raise StateMethodError(f"Irrelevant operation on the state: {self!r}")

    async def run(self) -> State:
        self.assert_not_obsolete()
        raise StateMethodError(f"Irrelevant operation on the state: {self!r}")

    async def finish(self) -> State:
        self.assert_not_obsolete()
        raise StateMethodError(f"Irrelevant operation on the state: {self!r}")

    async def reset(self, *_, **__) -> State:
        self.assert_not_obsolete()
        raise StateMethodError(f"Irrelevant operation on the state: {self!r}")

    async def close(self) -> State:
        self.assert_not_obsolete()
        raise StateMethodError(f"Irrelevant operation on the state: {self!r}")

    def interrupt(self) -> None:
        raise StateMethodError(f"Irrelevant operation on the state: {self!r}")

    def terminate(self) -> None:
        raise StateMethodError(f"Irrelevant operation on the state: {self!r}")

    def kill(self) -> None:
        raise StateMethodError(f"Irrelevant operation on the state: {self!r}")

    def exception(self) -> Optional[BaseException]:
        raise StateMethodError(f"Irrelevant operation on the state: {self!r}")

    def result(self) -> Any:
        raise StateMethodError(f"Irrelevant operation on the state: {self!r}")


class Created(State):
    """The state "created" """

    name = "created"

    def __init__(self, context: Context):
        self._context = context

    async def initialize(self) -> Initialized:
        self.assert_not_obsolete()
        next = await Initialized.create(self)
        self.obsolete()
        return next

    async def close(self) -> Closed:
        self.assert_not_obsolete()
        next = await Closed.create(self)
        self.obsolete()
        return next


class Initialized(State):
    """The state "initialized", ready to run"""

    name = "initialized"

    @classmethod
    async def create(cls, prev: State):
        self = cls(prev)
        await self._context.initialize(self)
        return self

    def __init__(self, prev: State):
        self._context = prev._context

    async def run(self) -> Running:
        self.assert_not_obsolete()
        next = await Running.create(self)
        self.obsolete()
        return next

    async def reset(self, *args, **kwargs) -> Initialized:
        self.assert_not_obsolete()
        await self._context.reset(*args, **kwargs)
        next = await Initialized.create(self)
        self.obsolete()
        return next

    async def close(self) -> Closed:
        self.assert_not_obsolete()
        next = await Closed.create(self)
        self.obsolete()
        return next


class Running(State):
    """The state "running", the script is being executed."""

    name = "running"

    @classmethod
    async def create(cls, prev: State):
        self = cls(prev)
        await self._context.run(self)
        return self

    def __init__(self, prev: State):
        self._context = prev._context

    async def finish(self) -> Finished:
        self.assert_not_obsolete()
        next = await Finished.create(self)
        self.obsolete()
        return next

    def interrupt(self) -> None:
        self._context.interrupt()

    def terminate(self) -> None:
        self._context.terminate()

    def kill(self) -> None:
        self._context.kill()


class Finished(State):
    """The state "finished", the script execution has finished"""

    name = "finished"

    @classmethod
    async def create(cls, prev: State):
        self = cls(prev)
        await self._context.finish(self)
        return self

    def __init__(self, prev: State):
        self._context = prev._context

    def exception(self) -> Optional[BaseException]:
        """Return the exception of the script execution

        Return None if no exception has been raised.

        """
        return self._context.exception

    def result(self) -> Any:
        """Return the result of the script execution

        None in the current implementation as the build-in function
        exec() returns None.

        Re-raise the exception if an exception has been raised in the
        script execution.

        """

        if exc := self._context.exception:
            raise exc

        return self._context.result

    async def finish(self) -> Finished:
        self.assert_not_obsolete()
        return self

    async def reset(self, *args, **kwargs) -> Initialized:
        self.assert_not_obsolete()
        await self._context.reset(*args, **kwargs)
        next = await Initialized.create(self)
        self.obsolete()
        return next

    async def close(self) -> Closed:
        self.assert_not_obsolete()
        next = await Closed.create(self)
        self.obsolete()
        return next


class Closed(State):
    """The state "closed" """

    name = "closed"

    @classmethod
    async def create(cls, prev: State):
        self = cls(prev)
        await self._context.close(self)
        return self

    def __init__(self, prev: State):
        self._context = prev._context

    async def close(self) -> Closed:
        self.assert_not_obsolete()
        return self



import asyncio
from functools import partial

import pytest

from nextline.utils import ExcThread, to_thread
from nextline.utils import ThreadTaskIdComposer as IdComposer
from nextline.types import ThreadTaskId, ThreadNo, TaskNo


def assert_call(obj: IdComposer, expected: ThreadTaskId, has_id: bool = False):
    assert obj.has_id() is has_id
    assert obj.has_id() is has_id
    assert expected == obj()
    assert expected == obj()
    assert obj.has_id()


async def async_assert_call(
    obj: IdComposer, expected: ThreadTaskId, has_id: bool = False
):
    assert obj.has_id() is has_id
    assert obj.has_id() is has_id
    await asyncio.sleep(0)
    assert expected == obj()
    await asyncio.sleep(0)
    assert expected == obj()
    assert obj.has_id()


@pytest.fixture()
def obj():
    y = IdComposer()
    yield y


def test_compose(obj: IdComposer):
    expected = ThreadTaskId(ThreadNo(1), None)
    assert_call(obj, expected)

    obj.reset()

    assert_call(obj, expected, True)


def test_threads(obj: IdComposer):
    expected = ThreadTaskId(ThreadNo(1), None)
    assert_call(obj, expected)

    expected = ThreadTaskId(ThreadNo(2), None)
    t = ExcThread(target=assert_call, args=(obj, expected))
    t.start()
    t.join()

    expected = ThreadTaskId(ThreadNo(3), None)
    t = ExcThread(target=assert_call, args=(obj, expected))
    t.start()
    t.join()

    obj.reset()

    expected = ThreadTaskId(ThreadNo(1), None)
    assert_call(obj, expected, True)

    expected = ThreadTaskId(ThreadNo(1), None)
    t = ExcThread(target=assert_call, args=(obj, expected))
    t.start()
    t.join()

    expected = ThreadTaskId(ThreadNo(2), None)
    t = ExcThread(target=assert_call, args=(obj, expected))
    t.start()
    t.join()


async def test_async_coroutine(obj: IdComposer):
    expected = ThreadTaskId(ThreadNo(1), TaskNo(1))
    assert_call(obj, expected)

    # run in the same task
    await async_assert_call(obj, expected, True)
    await async_assert_call(obj, expected, True)

    obj.reset()

    await async_assert_call(obj, expected, True)
    await async_assert_call(obj, expected, True)


async def test_async_tasks(obj: IdComposer):
    expected = ThreadTaskId(ThreadNo(1), TaskNo(1))
    assert_call(obj, expected)

    expected = ThreadTaskId(ThreadNo(1), TaskNo(2))
    t = asyncio.create_task(async_assert_call(obj, expected))
    await t

    expected = ThreadTaskId(ThreadNo(1), TaskNo(3))
    t = asyncio.create_task(async_assert_call(obj, expected))
    await t

    obj.reset()

    expected = ThreadTaskId(
        ThreadNo(1), TaskNo(1)
    )  # the old id is still there
    assert_call(obj, expected, True)

    expected = ThreadTaskId(ThreadNo(1), TaskNo(1))  # task no is reset to 1
    t = asyncio.create_task(async_assert_call(obj, expected))
    await t

    expected = ThreadTaskId(ThreadNo(1), TaskNo(2))
    t = asyncio.create_task(async_assert_call(obj, expected))
    await t


async def test_async_tasks_gather(obj: IdComposer):
    expected = ThreadTaskId(ThreadNo(1), TaskNo(1))
    assert_call(obj, expected)

    expected = ThreadTaskId(ThreadNo(1), TaskNo(2))
    t1 = asyncio.create_task(async_assert_call(obj, expected))
    expected = ThreadTaskId(ThreadNo(1), TaskNo(3))
    t2 = asyncio.create_task(async_assert_call(obj, expected))
    aws = {t1, t2}
    await asyncio.gather(*aws)

    obj.reset()

    expected = ThreadTaskId(ThreadNo(1), TaskNo(1))
    assert_call(obj, expected, True)

    expected = ThreadTaskId(ThreadNo(1), TaskNo(1))
    t1 = asyncio.create_task(async_assert_call(obj, expected))
    expected = ThreadTaskId(ThreadNo(1), TaskNo(2))
    t2 = asyncio.create_task(async_assert_call(obj, expected))
    aws = {t1, t2}
    await asyncio.gather(*aws)


def test_async_asyncio_run(obj: IdComposer):
    expected = ThreadTaskId(ThreadNo(1), None)
    assert_call(obj, expected)

    expected = ThreadTaskId(ThreadNo(1), TaskNo(1))
    asyncio.run(async_assert_call(obj, expected))

    obj.reset()

    expected = ThreadTaskId(ThreadNo(1), None)
    assert_call(obj, expected, True)

    expected = ThreadTaskId(ThreadNo(1), TaskNo(1))
    asyncio.run(async_assert_call(obj, expected))


async def test_async_asyncio_to_thread(obj: IdComposer):
    expected = ThreadTaskId(ThreadNo(1), TaskNo(1))
    assert_call(obj, expected)

    expected = ThreadTaskId(ThreadNo(2), None)
    await to_thread(partial(assert_call, obj, expected))

    expected = ThreadTaskId(
        ThreadNo(2), None
    )  # to_thread uses the same thread
    await to_thread(partial(assert_call, obj, expected, True))

    obj.reset()

    expected = ThreadTaskId(ThreadNo(1), TaskNo(1))
    assert_call(obj, expected, True)

    expected = ThreadTaskId(
        ThreadNo(2), None
    )  # to_thread uses the same thread
    await to_thread(partial(assert_call, obj, expected, True))

    expected = ThreadTaskId(
        ThreadNo(2), None
    )  # to_thread uses the same thread
    await to_thread(partial(assert_call, obj, expected, True))


async def async_nested(obj: IdComposer, expected_thread_id):
    expected1 = ThreadTaskId(expected_thread_id, TaskNo(1))
    assert_call(obj, expected1)
    await async_assert_call(obj, expected1, True)

    expected2 = ThreadTaskId(expected_thread_id, TaskNo(2))
    t1 = asyncio.create_task(async_assert_call(obj, expected2))
    expected3 = ThreadTaskId(expected_thread_id, TaskNo(3))
    t2 = asyncio.create_task(async_assert_call(obj, expected3))
    aws = {t1, t2}
    await asyncio.gather(*aws)


def nested(obj: IdComposer, expected_thread_id):
    expected = ThreadTaskId(expected_thread_id, None)
    assert_call(obj, expected)

    asyncio.run(async_nested(obj, expected_thread_id))


def test_nested(obj: IdComposer):
    expected = ThreadTaskId(ThreadNo(1), None)
    assert_call(obj, expected)

    expected_thread_id = 2
    t = ExcThread(target=nested, args=(obj, expected_thread_id))
    t.start()
    t.join()

    expected_thread_id = 3
    t = ExcThread(target=nested, args=(obj, expected_thread_id))
    t.start()
    t.join()

    obj.reset()

    expected = ThreadTaskId(ThreadNo(1), None)
    assert_call(obj, expected, True)

    expected_thread_id = 1  # reset to 1
    t = ExcThread(target=nested, args=(obj, expected_thread_id))
    t.start()
    t.join()

    expected_thread_id = 2
    t = ExcThread(target=nested, args=(obj, expected_thread_id))
    t.start()
    t.join()



from __future__ import annotations

from concurrent.futures import Executor, ProcessPoolExecutor
from concurrent.futures.process import BrokenProcessPool
from multiprocessing import Process

import os
import signal
import asyncio

from functools import partial
from typing import Callable, Generic, Optional, TypeVar
from typing_extensions import ParamSpec


_T = TypeVar("_T")
_P = ParamSpec("_P")


async def run_in_process(
    executor_factory: Optional[Callable[[], Executor]],
    func: Callable[_P, _T],
    *func_args: _P.args,
    **func_kwargs: _P.kwargs,
) -> RunInProcess[_T, _P]:
    if executor_factory is None:
        executor_factory = partial(ProcessPoolExecutor, max_workers=1)
    return await RunInProcess.create(
        executor_factory, func, *func_args, **func_kwargs
    )


class RunInProcess(Generic[_T, _P]):
    @classmethod
    async def create(
        cls,
        executor_factory: Callable[[], Executor],
        func: Callable[_P, _T],
        *func_args: _P.args,
        **func_kwargs: _P.kwargs,
    ):
        self = cls(executor_factory, func, *func_args, **func_kwargs)
        assert await self._event.wait()
        return self

    def __init__(
        self,
        executor_factory: Callable[[], Executor],
        func: Callable[_P, _T],
        *func_args: _P.args,
        **func_kwargs: _P.kwargs,
    ):
        self._executor_factory = executor_factory
        self._func_call = partial(func, *func_args, **func_kwargs)
        self._event = asyncio.Event()
        self._task = asyncio.create_task(self._run())
        self._process: Optional[Process] = None
        self._future: Optional[asyncio.Future] = None
        self._exc: Optional[BaseException] = None

    def __repr__(self):
        return f"<{self.__class__.__name__} {self._process!r} {self._future}>"

    async def _run(self) -> Optional[_T]:

        with self._executor_factory() as executor:
            loop = asyncio.get_running_loop()
            self._future = loop.run_in_executor(executor, self._func_call)
            if isinstance(executor, ProcessPoolExecutor):
                self._process = list(executor._processes.values())[0]
            self._event.set()
            try:
                return await self._future
            except BrokenProcessPool:
                # NOTE: Not possible to use "as" for unknown reason.
                return None
            except BaseException as e:
                self._exc = e
                return None

    def interrupt(self) -> None:
        if self._process and self._process.pid:
            os.kill(self._process.pid, signal.SIGINT)

    def terminate(self) -> None:
        if self._process:
            self._process.terminate()

    def kill(self) -> None:
        if self._process:
            self._process.kill()

    def __await__(self):
        # NOTE: this method can be as simple as the following one line if it
        # only awaits for the task:
        #
        # return self._task.__await__()
        #
        # "yield from" is used to execute extra code.
        # https://stackoverflow.com/a/48261042/7309855

        ret = yield from self._task.__await__()
        if self._exc:
            raise self._exc
        return ret



import asyncio
from threading import Thread
from itertools import permutations

import pytest
from unittest.mock import Mock

from typing import List, MutableSequence, Sequence, Tuple, Set, Any

from nextline.process.call import call_with_trace
from nextline.process.trace import TraceDispatchThreadOrTask


@pytest.fixture()
def created() -> List[Mock]:
    """List of mock objects created by the factory fixture"""
    return []


@pytest.fixture()
def factory(created: MutableSequence[Mock]) -> Mock:
    """Function that creates a Mock object that returns itself

    Created Mock objects are used as trace functions.

    The fixture "created" contains the created objects.
    """

    def side_effect():
        trace_func = Mock()
        trace_func.return_value = trace_func
        created.append(trace_func)
        return trace_func

    y = Mock(side_effect=side_effect)
    return y


def f_pass():
    pass


def f_call():
    return f_pass()


def f_thread():
    t = Thread(target=f_call)
    t.start()
    t.join()


async def a_sleep():
    await asyncio.sleep(0)
    await asyncio.sleep(0.001)


def a_run():
    asyncio.run(a_sleep())


async def a_task():
    t1 = asyncio.create_task(a_sleep())
    t2 = asyncio.create_task(a_sleep())
    await asyncio.gather(t1, t2)


def a_run_task():
    asyncio.run(a_task())


def f_thread_run_task():
    t = Thread(target=a_run_task)
    t.start()
    t.join()


params = [
    pytest.param(
        f_pass,
        [{(__name__, f_pass.__name__)}],
    ),
    pytest.param(
        f_call,
        [{(__name__, f_call.__name__), (__name__, f_pass.__name__)}],
    ),
    pytest.param(
        f_thread,
        [
            {(__name__, f_thread.__name__)},
            {(__name__, f_call.__name__), (__name__, f_pass.__name__)},
        ],
    ),
    pytest.param(
        a_run,
        [{(__name__, a_run.__name__)}, {(__name__, a_sleep.__name__)}],
    ),
    pytest.param(
        a_run_task,
        [
            {(__name__, a_run_task.__name__)},
            {(__name__, a_task.__name__)},
            {(__name__, a_sleep.__name__)},
            {(__name__, a_sleep.__name__)},
        ],
    ),
    pytest.param(
        f_thread_run_task,
        [
            {(__name__, f_thread_run_task.__name__)},
            {(__name__, a_run_task.__name__)},
            {(__name__, a_task.__name__)},
            {(__name__, a_sleep.__name__)},
            {(__name__, a_sleep.__name__)},
        ],
    ),
]


@pytest.mark.parametrize("func, expected", params)
def test_one(factory: Mock, created: MutableSequence[Mock], func, expected):
    obj = TraceDispatchThreadOrTask(factory=factory)
    call_with_trace(func, obj)
    print()
    # print(traced(created))
    actual = traced(created)
    # print(actual)
    # print(len(actual))
    assert is_unordered_list_of_subsets(expected, actual)


def traced(created: Sequence[Mock]) -> List[Set[Tuple[str, str]]]:
    """A list of sets, each with tuples of traced module and func names

    e.g., [{("module", "func"), ...}, ...]
    """
    args_list = [
        [call.args for call in mock.call_args_list] for mock in created
    ]
    return [
        {
            (frame.f_globals.get("__name__"), frame.f_code.co_name)
            for frame, event, _ in args
            if event == "return"
        }
        for args in args_list
    ]


def is_list_of_subsets(
    sets1: Sequence[Set[Any]], sets2: Sequence[Set[Any]]
) -> bool:
    if not len(sets1) <= len(sets2):
        return False
    intersections = [s1 & s2 for s1, s2 in zip(sets1, sets2)]
    return sets1 == intersections


def is_unordered_list_of_subsets(
    sets1: Sequence[Set[Any]], sets2: Sequence[Set[Any]]
) -> bool:
    for p in permutations(sets2):
        if is_list_of_subsets(sets1, p):
            return True
    return False


params = [
    pytest.param([], [], True),
    pytest.param([], [set()], True),
    pytest.param([set()], [], False),
    pytest.param([set()], [set()], True),
    pytest.param([{1}], [{1}], True),
    pytest.param([{1}], [{1}, {2}], True),
    pytest.param([{1}], [{1, 2}], True),
    pytest.param([{1}], [{2}], False),
    pytest.param([{1}, {2}], [{1, 10}, {2, 20}], True),
    pytest.param([{1}, {2}], [{2, 20}, {1, 10}], True),
    pytest.param([{1}, {2}], [{1, 10}, {3, 30}], False),
    pytest.param([{1}, {2}], [{3, 30}, {1, 10}], False),
    pytest.param([{1}, {1}, {2}], [{1, 10}, {2, 20}], False),
    pytest.param([{1}, {1}, {2}], [{1, 10}, {1, 30}, {2, 20}], True),
]


@pytest.mark.parametrize("sets1, sets2, expected", params)
def test_is_unordered_list_of_subset(
    sets1: Sequence[Set[Any]], sets2: Sequence[Set[Any]], expected: bool
):
    assert is_unordered_list_of_subsets(sets1, sets2) is expected



from __future__ import annotations

import time
import signal
from functools import partial
from concurrent.futures import Executor, ProcessPoolExecutor
from multiprocessing import Event
from types import FrameType
from typing import TYPE_CHECKING, Callable, NoReturn, Optional

import pytest

from nextline.utils import run_in_process

if TYPE_CHECKING:
    from multiprocessing import _EventType


_event: Optional[_EventType] = None


def initializer(event: _EventType) -> None:
    global _event
    _event = event


class MockError(Exception):
    pass


class Handled(Exception):
    pass


def func() -> str:
    return "foo"


def func_raise() -> NoReturn:
    raise MockError()


def func_slow() -> NoReturn:
    assert _event
    _event.set()
    time.sleep(10)
    raise RuntimeError("to be terminated by here")


def func_slow_catch() -> str:
    assert _event
    try:
        _event.set()
        time.sleep(10)
    except KeyboardInterrupt:
        return "foo"
    return "bar"


def handler(signum: signal._SIGNUM, frame: FrameType):
    raise Handled


def func_slow_handle() -> str:
    assert _event
    signal.signal(signal.SIGTERM, handler)  # type: ignore
    try:
        _event.set()
        time.sleep(10)
    except Handled:
        return "foo"
    return "bar"


async def test_one(executor_factory: Callable[[], Executor]) -> None:
    r = await run_in_process(executor_factory, func)
    assert "foo" == await r
    assert r._process
    assert r._process.exitcode == 0


async def test_default_executor() -> None:
    r = await run_in_process(None, func)
    assert "foo" == await r
    assert r._process
    assert r._process.exitcode == 0


async def test_repr(executor_factory: Callable[[], Executor]) -> None:
    r = await run_in_process(executor_factory, func)
    repr(r)
    await r


async def test_error(executor_factory: Callable[[], Executor]) -> None:
    r = await run_in_process(executor_factory, func_raise)
    with pytest.raises(MockError):
        await r
    assert r._process
    assert r._process.exitcode == 0


async def test_interrupt(
    executor_factory: Callable[[], Executor], event: _EventType
) -> None:
    r = await run_in_process(executor_factory, func_slow)
    event.wait()
    r.interrupt()
    with pytest.raises(KeyboardInterrupt):
        await r
    assert r._process
    assert r._process.exitcode == 0


async def test_interrupt_catch(
    executor_factory: Callable[[], Executor], event: _EventType
) -> None:
    r = await run_in_process(executor_factory, func_slow_catch)
    event.wait()
    r.interrupt()
    assert "foo" == await r
    assert r._process
    assert r._process.exitcode == 0


async def test_terminate(
    executor_factory: Callable[[], Executor], event: _EventType
) -> None:
    r = await run_in_process(executor_factory, func_slow)
    event.wait()
    r.terminate()
    assert None is await r
    assert r._process
    assert r._process.exitcode == -signal.SIGTERM


async def test_terminate_handle(
    executor_factory: Callable[[], Executor], event: _EventType
) -> None:
    r = await run_in_process(executor_factory, func_slow_handle)
    event.wait()
    r.terminate()
    assert "foo" == await r
    assert r._process
    assert r._process.exitcode == 0


async def test_kill(
    executor_factory: Callable[[], Executor], event: _EventType
) -> None:
    r = await run_in_process(executor_factory, func_slow)
    event.wait()
    r.kill()
    assert None is await r
    assert r._process
    assert r._process.exitcode == -signal.SIGKILL


@pytest.fixture
def executor_factory(event: _EventType) -> Callable[[], Executor]:
    return partial(
        ProcessPoolExecutor,
        max_workers=1,
        initializer=initializer,
        initargs=(event,),
    )


@pytest.fixture
def event() -> _EventType:
    return Event()



from __future__ import annotations

import enum
from asyncio import Queue, Condition

from typing import (
    AsyncIterator,
    Generic,
    Literal,
    Optional,
    List,
    Tuple,
    TypeVar,
)


class _M(enum.Enum):
    # TODO: Using Enum as sentinel for now as suggested in
    # https://stackoverflow.com/a/60605919/7309855. It still has a problem. For
    # example, the type of yielded values in subscribe() is not correctly
    # inferred as _T.

    START = object()
    END = object()


_T = TypeVar("_T")


class PubSubItem(Generic[_T]):
    """Distribute items to subscribers

    A new subscriber immediately receives the latest item and then wait for new
    items.

    The order of the items is preserved.
    """

    def __init__(self):

        self._qs_out: List[Queue[Tuple[int, _T | Literal[_M.END]]]] = []
        self._last_enumerated: Tuple[
            int, _T | Literal[_M.START] | Literal[_M.END]
        ] = (-1, _M.START)

        self._last_item: _T | Literal[_M.START] = _M.START
        self._idx = -1

        self._closed: bool = False
        self._lock_close = Condition()

    @property
    def nsubscriptions(self) -> int:
        """The number of the subscribers"""
        return len(self._qs_out)

    async def publish(self, item: _T) -> None:
        """Send data to subscribers"""
        if self._closed:
            raise RuntimeError(f"{self} is closed.")
        self._last_item = item
        await self._publish(item)

    def latest(self) -> _T:
        """Most recent data that have been published"""
        if self._last_item is _M.START:
            raise LookupError
        return self._last_item

    async def subscribe(
        self,
        last: Optional[bool] = True,
    ) -> AsyncIterator[_T]:
        """Yield data as they are put

        If `last` is true, yield immediately the most recent data before
        waiting for new data.
        """
        q: Queue[Tuple[int, _T | Literal[_M.END]]] = Queue()

        self._qs_out.append(q)

        last_idx, last_item = self._last_enumerated

        if last_item is _M.END:
            return

        try:
            if last and last_item is not _M.START:
                yield last_item

            while True:
                idx, item = await q.get()
                if item is _M.END:
                    break
                if last_idx < idx:
                    yield item

        finally:
            self._qs_out.remove(q)

    async def close(self) -> None:
        """End gracefully"""
        async with self._lock_close:
            if self._closed:
                return
            self._closed = True
            await self._publish(_M.END)

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        del exc_type, exc_value, traceback
        await self.close()

    async def _publish(self, item: _T | Literal[_M.END]) -> None:
        self._idx += 1
        self._last_enumerated = (self._idx, item)
        for q in list(self._qs_out):  # list in case it changes
            await q.put(self._last_enumerated)



from functools import partial

import pytest
from unittest.mock import Mock

from typing import Set

from nextline.process.trace import TraceAddFirstModule

from .funcs import TraceSummary

from . import module_a


def test_one(
    target: TraceSummary,
    probe: TraceSummary,
    ref: TraceSummary,
    modules_to_trace: Set[str],
    modules_to_trace_init: Set[str],
    module_name: str,
):
    assert ref.call.module
    assert ref.return_.module
    assert ref.call == target.call
    assert not target.return_.module
    assert ref == probe

    assert modules_to_trace_init is not modules_to_trace
    assert {module_name} | modules_to_trace_init == modules_to_trace


def f1():
    module_a.func_a()


f2 = partial(f1)

statement = """
module_a.func_a()
"""

code = compile(statement, "<string>", "exec")
globals_ = {"module_a": module_a}  # without __name__
f3 = partial(exec, code, globals_)


@pytest.fixture(
    params=[
        (f1, __name__),
        (f2, __name__),
        (f3, module_a.__name__),
    ]
)
def func_and_module_name(request):
    yield request.param


@pytest.fixture()
def func(func_and_module_name):
    yield func_and_module_name[0]


@pytest.fixture()
def module_name(func_and_module_name):
    yield func_and_module_name[1]


@pytest.fixture(params=[set(), {"some_module"}])
def modules_to_trace_init(request):
    y = request.param
    yield y


@pytest.fixture()
def modules_to_trace(modules_to_trace_init):
    y = set(modules_to_trace_init)
    yield y


@pytest.fixture()
def target_trace_func(probe_trace_func: Mock, modules_to_trace: Set[str]):
    y = TraceAddFirstModule(
        trace=probe_trace_func,
        modules_to_trace=modules_to_trace,
    )
    yield y


@pytest.fixture()
def thread():
    yield False



from __future__ import annotations

import pytest
from unittest.mock import Mock

from typing import TYPE_CHECKING, Callable, Any, Set

from nextline.process.call import call_with_trace

if TYPE_CHECKING:
    from sys import _TraceFunc as TraceFunc

from .funcs import TraceSummary, summarize_trace_calls


@pytest.fixture()
def target(
    wrap_target_trace_func: Mock,
    modules_in_summary: Set[str] | None,
    run_target,
) -> TraceSummary:
    _ = run_target
    y = summarize_trace_calls(
        wrap_target_trace_func, modules=modules_in_summary
    )
    return y


@pytest.fixture()
def probe(
    probe_trace_func: Mock,
    modules_in_summary: Set[str] | None,
    run_target,
) -> TraceSummary:
    _ = run_target
    y = summarize_trace_calls(probe_trace_func, modules=modules_in_summary)
    return y


@pytest.fixture()
def ref(
    ref_trace_func: Mock,
    modules_in_summary: Set[str] | None,
    run_ref,
) -> TraceSummary:
    _ = run_ref
    y = summarize_trace_calls(ref_trace_func, modules=modules_in_summary)
    return y


@pytest.fixture()
def modules_in_summary() -> Set[str] | None:
    return None


@pytest.fixture()
def run_target(
    func: Callable[[], Any],
    wrap_target_trace_func: Mock,
    thread: bool,
) -> None:
    call_with_trace(func=func, trace=wrap_target_trace_func, thread=thread)
    return


@pytest.fixture()
def run_ref(
    func: Callable[[], Any],
    ref_trace_func: Mock,
    thread: bool,
) -> None:
    call_with_trace(func=func, trace=ref_trace_func, thread=thread)
    return


@pytest.fixture()
def wrap_target_trace_func(target_trace_func: TraceFunc) -> Mock:
    wrap = Mock(wraps=target_trace_func)

    def side_effect(*a, **k):
        local_trace_func = target_trace_func(*a, **k)
        if local_trace_func is target_trace_func:
            return wrap
        return local_trace_func

    wrap.side_effect = side_effect
    return wrap


@pytest.fixture()
def probe_trace_func() -> Mock:
    y = Mock()
    y.return_value = y
    return y


@pytest.fixture()
def ref_trace_func() -> Mock:
    y = Mock()
    y.return_value = y
    return y



import time
import random
from threading import current_thread

from nextline.utils import ThreadDoneCallback
from nextline.utils import ExcThread

import pytest
from unittest.mock import Mock


def target(obj: ThreadDoneCallback):
    """To run in a thread"""
    assert current_thread() == obj.register()
    delay = random.random() * 0.01
    time.sleep(delay)


class Done:
    """A callback function"""

    def __init__(self):
        self.args = set()

    def __call__(self, arg):
        self.args.add(arg)


@pytest.fixture()
def done():
    """A callback function"""
    yield Done()


def test_close(done: Done):
    obj = ThreadDoneCallback(done=done)
    t = ExcThread(target=target, args=(obj,))
    t.start()
    time.sleep(0.005)
    obj.close()
    assert {t} == done.args
    t.join()


def test_with(done: Done):
    with ThreadDoneCallback(done=done) as obj:
        t = ExcThread(target=target, args=(obj,))
        t.start()
        time.sleep(0.005)
    assert {t} == done.args
    t.join()


def test_register_arg(done: Done):
    def target():
        delay = random.random() * 0.01
        time.sleep(delay)

    with ThreadDoneCallback(done=done) as obj:
        t = ExcThread(target=target)

        # manually provide the thread object
        assert t == obj.register(t)

        t.start()
        time.sleep(0.005)

    assert {t} == done.args
    t.join()


def test_daemon(done: Done):
    """Not blocked even if close() is not called"""
    obj = ThreadDoneCallback(done=done)
    t = ExcThread(target=target, args=(obj,))
    t.start()
    time.sleep(0.005)
    # obj.close()  # not call closed()
    t.join()


@pytest.mark.parametrize("n_threads", [0, 1, 2, 5, 10])
def test_multiple(n_threads: int, done: Done):

    with ThreadDoneCallback(done=done) as obj:

        threads = {
            ExcThread(target=target, args=(obj,)) for _ in range(n_threads)
        }
        for t in threads:
            t.start()
        time.sleep(0.005)

    assert threads == done.args

    for t in threads:
        t.join()


def test_raise_close_from_thread(done: Done):
    def target(obj: ThreadDoneCallback):
        obj.register()

        obj.close()
        # raised originally here, then re-raised at the join() by
        # ExcThread in the main thread.

    obj = ThreadDoneCallback(done=done)
    t = ExcThread(target=target, args=(obj,))
    t.start()
    with pytest.raises(RuntimeError):
        t.join()


def test_raise_in_done():
    done = Mock(side_effect=ValueError)
    obj = ThreadDoneCallback(done=done)
    t = ExcThread(target=target, args=(obj,))
    t.start()
    time.sleep(0.005)

    with pytest.raises(ValueError):
        obj.close()

    t.join()


def test_interval(done: Done):
    interval = 0.02
    obj = ThreadDoneCallback(done=done, interval=interval)
    t = ExcThread(target=target, args=(obj,))
    t.start()
    time.sleep(0.005)

    obj.close()
    t.join()


def test_done_none():
    with ThreadDoneCallback() as obj:
        t = ExcThread(target=target, args=(obj,))
        t.start()
        time.sleep(0.005)
    assert not t.is_alive()
    t.join()



from __future__ import annotations

import pytest
from unittest.mock import Mock, call

from nextline.state import State, Closed

from .base import BaseTestState


class TestClosed(BaseTestState):

    state_class = Closed

    @pytest.fixture()
    def state(self, closed: State) -> State:
        return closed

    def test_state(self, state: State, context: Mock):
        super().test_state(state, context)
        assert [call(state)] == context.close.call_args_list

    async def test_close(self, state: State):
        # The same object should be returned no matter
        # how many times called.
        assert state is await state.close()
        assert state is await state.close()
        assert state is await state.close()
        assert "obsolete" not in repr(state)



from __future__ import annotations
from io import TextIOWrapper
from queue import Queue


class StreamOut(TextIOWrapper):
    def __init__(self, queue: Queue[str]):
        self.queue = queue

    def write(self, s: str, /) -> int:
        self.queue.put(s)
        return len(s)

    def flush(self):
        pass


class StreamIn(TextIOWrapper):
    def __init__(self, queue: Queue[str]):
        self.queue = queue

    def readline(self):
        return self.queue.get()



from __future__ import annotations

import sys
import os
import signal

from pdb import Pdb

from multiprocessing import Process, Queue
from time import sleep

import pytest

from nextline.process.pdb.stream import StreamIn, StreamOut


def f():
    sleep(0.001)


def run(queue_stdin: Queue[str], queue_stdout: Queue[str]) -> None:
    pdb = Pdb(
        stdin=StreamIn(queue_stdin),
        stdout=StreamOut(queue_stdout),
        readrc=False,
    )
    pdb.reset()
    trace_org = sys.gettrace()
    sys.settrace(pdb.trace_dispatch)
    try:
        f()
    finally:
        sys.settrace(trace_org)
        queue_stdout.put(None)  # type: ignore


def test_one(queue_stdin: Queue[str], queue_stdout: Queue[str]) -> None:

    p = Process(target=run, args=(queue_stdin, queue_stdout))
    p.start()

    print()
    first = True
    while m := queue_stdout.get():
        print(m, end="")
        if m != "(Pdb) ":
            continue
        if first and p.pid:
            os.kill(p.pid, signal.SIGINT)
            first = False
            continue
        command = "next"
        print(command)
        queue_stdin.put(command)

    p.join()


@pytest.fixture
def queue_stdin() -> Queue[str]:
    y: Queue[str] = Queue()
    return y


@pytest.fixture
def queue_stdout() -> Queue[str]:
    y: Queue[str] = Queue()
    return y



import pytest
from unittest.mock import Mock

from nextline.process.trace import TraceSkipLambda

from .funcs import TraceSummary

from . import module_a


def test_one(target: TraceSummary, probe: TraceSummary, ref: TraceSummary):
    assert ref.call.func
    assert ref.return_.func
    assert ref.call.func == target.call.func
    assert not target.return_.func
    assert set(ref.call.func) - {"<lambda>"} == set(probe.call.func)
    assert set(ref.return_.func) - {"<lambda>"} == set(probe.return_.func)


def f():
    module_a.func_a()


def g():
    (lambda: module_a.func_a())()


@pytest.fixture(params=[f, g, lambda: module_a.func_a()])
def func(request):
    yield request.param


@pytest.fixture()
def target_trace_func(probe_trace_func: Mock):
    y = TraceSkipLambda(trace=probe_trace_func)
    yield y


@pytest.fixture()
def thread():
    yield False



from __future__ import annotations

from functools import partial
from pdb import Pdb

from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from .proxy import PdbInterface


def getlines(func_org, statement, filename, module_globals=None):
    if filename == "<string>":
        return statement.split("\n")
    return func_org(filename, module_globals)


class CustomizedPdb(Pdb):
    """A Pdb subclass that calls back PdbProxy"""

    def __init__(self, pdbi: PdbInterface, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._pdbi = pdbi

        # self.quitting = True # not sure if necessary

        # stop at the first line
        self.botframe = None
        self._set_stopinfo(None, None)  # type: ignore

    def trace_dispatch(self, frame, event, arg):
        """The main trace function of Bdb"""
        return super().trace_dispatch(frame, event, arg)

    def _cmdloop(self):
        """Prompt user input"""
        try:
            self._pdbi.entering_cmdloop()
        except RuntimeError:
            # "step" at the last line
            # https://github.com/simonsobs/nextline/issues/1
            return
        try:
            # super()._cmdloop()  # This catches KeyboardInterrupt
            super().cmdloop()
        finally:
            self._pdbi.exited_cmdloop()

    def set_continue(self):
        """override bdb.set_continue()

        To avoid sys.settrace(None) called in bdb.set_continue()

        """
        self._set_stopinfo(self.botframe, None, -1)

    # def break_here(self, frame):
    #     ret = super().break_here(frame)
    #     print('break_here', ret)
    #     return ret

    # def stop_here(self, frame):
    #     ret = super().stop_here(frame)
    #     print('stop_here', ret)
    #     return ret

    # def bp_commands(self, frame):
    #     return True

    def do_list(self, arg):
        statement = self._pdbi.statement
        import linecache

        getlines_org = linecache.getlines
        linecache.getlines = partial(getlines, getlines_org, statement)
        try:
            return super().do_list(arg)
        finally:
            linecache.getlines = getlines_org



from __future__ import annotations

import sys
import threading
import timeit


from nextline.process.trace import TraceSkipModule, MODULES_TO_SKIP
from nextline.utils import profile_func


def test_one():
    trace = TraceSkipModule(
        trace=lambda *_: None,
        skip=MODULES_TO_SKIP,
    )
    frame = sys._current_frames()[threading.current_thread().ident]
    # t = timeit.timeit(lambda: trace(frame, "line", None), number=20_000)
    t = timeit.timeit(lambda: trace(frame, "line", None), number=200_000)
    # t = timeit.timeit(lambda: trace(frame, "line", None), number=2_000_000)
    # print()
    # print(t)

    def func():
        for _ in range(20_000):
            trace(frame, "line", None)

    profile, _ = profile_func(func)
    # print(profile)



import pytest

from nextline.utils import ExcThread


class ErrorInThread(Exception):
    pass


def func_raise():
    raise ErrorInThread()


def test_return():
    t = ExcThread(target=func_raise)
    t.start()
    with pytest.raises(ErrorInThread):
        t.join()



import sys
from threading import Thread
import traceback

import pytest
from unittest.mock import Mock

from nextline.process.call import call_with_trace


@pytest.fixture()
def trace():
    f = Mock()
    f.return_value = f
    yield f


def test_simple(trace):
    def func():
        x = 123
        return x

    trace_org = sys.gettrace()
    ret, exc = call_with_trace(func, trace=trace)
    assert trace_org == sys.gettrace()
    assert 123 == ret
    assert exc is None
    # print(trace.call_args_list)
    assert 4 == trace.call_count  # "call", "line", "line", "return"


class MockError(Exception):
    pass


def test_raise(trace):
    def func():
        raise MockError()

    trace_org = sys.gettrace()

    ret, exc = call_with_trace(func, trace=trace)

    assert trace_org == sys.gettrace()

    assert ret is None
    assert isinstance(exc, MockError)

    assert 4 == trace.call_count  # "call", "line", "exception", "return"

    # assert the frame of call_with_trace() is removed
    formatted = traceback.format_exception(type(exc), exc, exc.__traceback__)
    assert len(formatted) == 3


@pytest.mark.parametrize("thread", [True, False])
def test_threading(trace, thread):
    def f1():
        return

    def func():
        t1 = Thread(target=f1)
        t1.start()
        t1.join()

    trace_org = sys.gettrace()

    call_with_trace(func, trace=trace, thread=thread)

    assert trace_org == sys.gettrace()

    traced = {
        (
            c.args[0].f_globals.get("__name__"),
            c.args[0].f_code.co_name,
        )
        for c in trace.call_args_list
    }
    # set: {(<module name>, <func name>)}

    expected_subset = {(func.__module__, func.__name__)}

    if thread:
        expected_subset.update({(f1.__module__, f1.__name__)})

    assert expected_subset <= traced



from __future__ import annotations

import asyncio
import pytest

from nextline import Nextline

STATEMENT = """
import time

time.sleep(0.01)
""".lstrip()


async def test_run(nextline: Nextline):
    assert nextline.state == "initialized"

    await asyncio.gather(
        control(nextline),
        run(nextline),
    )
    assert nextline.state == "closed"


async def run(nextline: Nextline):
    await asyncio.sleep(0.01)
    await nextline.run()
    nextline.exception()
    nextline.result()
    await nextline.close()


async def control(nextline: Nextline):
    async for prompt_info in nextline.subscribe_prompt_info():
        if not prompt_info.open:
            continue
        nextline.send_pdb_command(
            "next", prompt_info.prompt_no, prompt_info.trace_no
        )


@pytest.fixture
async def nextline(statement):
    async with Nextline(statement) as y:
        yield y


@pytest.fixture
def statement():
    return STATEMENT



from __future__ import annotations

from queue import Queue  # noqa F401
from concurrent.futures import ThreadPoolExecutor
from contextlib import contextmanager
from logging import getLogger

from typing import Callable, Set, TypeVar, TypedDict, MutableMapping
from typing import Any, Tuple  # noqa F401
from typing_extensions import TypeAlias

from ..types import RunNo, TraceNo, PromptNo
from .trace import Trace
from .call import call_with_trace
from .callback import Callback, RegistrarProxy
from . import script

_T = TypeVar("_T")


PdbCommand: TypeAlias = str
QueueCommands: TypeAlias = "Queue[Tuple[PdbCommand, PromptNo, TraceNo] | None]"
QueueRegistry: TypeAlias = "Queue[Tuple[str, Any, bool]]"
PdbCiMap: TypeAlias = MutableMapping[
    TraceNo, Callable[[PdbCommand, PromptNo], Any]
]

_q_commands: QueueCommands | None = None
_q_registry: QueueRegistry | None = None


def set_queues(q_commands: QueueCommands, q_registry: QueueRegistry) -> None:
    global _q_commands, _q_registry
    _q_commands = q_commands
    _q_registry = q_registry


class RunArg(TypedDict):
    run_no: RunNo
    statement: str
    filename: str


class Context(TypedDict):
    callback: Callback
    pdb_ci_map: PdbCiMap
    modules_to_trace: Set[str]


def run(run_arg: RunArg) -> Tuple[Any, BaseException | None]:
    assert _q_registry
    assert _q_commands
    return run_(run_arg, _q_commands, _q_registry)


def run_(
    run_arg: RunArg,
    q_commands: QueueCommands,
    q_registry: QueueRegistry,
) -> Tuple[Any, BaseException | None]:

    run_no = run_arg["run_no"]
    statement = run_arg.get("statement")
    filename = run_arg.get("script_file_name", "<string>")

    try:
        code = _compile(statement, filename)
    except BaseException as e:
        return None, e

    pdb_ci_map: PdbCiMap = {}
    modules_to_trace: Set[str] = set()

    with Callback(
        run_no=run_no,
        registrar=RegistrarProxy(q_registry),
        modules_to_trace=modules_to_trace,
    ) as callback:

        context = Context(
            callback=callback,
            pdb_ci_map=pdb_ci_map,
            modules_to_trace=modules_to_trace,
        )

        trace = Trace(context=context)

        func = script.compose(code)

        with relay_commands(q_commands, pdb_ci_map):
            result, exception = call_with_trace(func, trace)

    return result, exception


def _compile(code, filename):
    if isinstance(code, str):
        code = compile(code, filename, "exec")
    return code


@contextmanager
def relay_commands(q_commands: QueueCommands, pdb_ci_map: PdbCiMap):
    def fn() -> None:
        assert q_commands
        while m := q_commands.get():
            command, prompt_no, trace_no = m
            pdb_ci = pdb_ci_map[trace_no]
            pdb_ci(command, prompt_no)

    with ThreadPoolExecutor(max_workers=1) as executor:
        future = executor.submit(try_again_on_error, fn)  # type: ignore
        try:
            yield
        finally:
            q_commands.put(None)
            future.result()


def try_again_on_error(func: Callable[[], _T]) -> _T:
    while True:
        try:
            return func()
        # except KeyboardInterrupt:
        #     raise
        except BaseException:
            logger = getLogger(__name__)
            logger.exception("")



from __future__ import annotations

import pytest
from unittest.mock import Mock, call, sentinel

from typing import Any

from nextline.state import State, Initialized, Finished, Closed

from .base import BaseTestState


class MockError(Exception):
    pass


class TestFinished(BaseTestState):

    state_class = Finished

    @pytest.fixture()
    def state(self, finished: State) -> State:
        return finished

    def test_state(self, state: State, context: Mock):
        super().test_state(state, context)
        assert [call(state)] == context.finish.call_args_list

    async def test_finish(self, state: State):
        # The same object should be returned no matter
        # how many times called.
        assert state is await state.finish()
        assert state is await state.finish()
        assert state is await state.finish()
        assert "obsolete" not in repr(state)

    async def test_reset(self, state: State, context: Mock):  # type: ignore
        reset = await state.reset(sentinel.args)
        assert isinstance(reset, Initialized)
        await self.assert_obsolete(state)
        assert [call(sentinel.args)] == context.reset.call_args_list

    async def test_close(self, state: State):
        closed = await state.close()
        assert isinstance(closed, Closed)
        await self.assert_obsolete(state)

    async def test_exception(self, state: State, mock_run_exception):  # type: ignore
        assert state.exception() is mock_run_exception

    async def test_result(self, state: State, mock_run_exception):  # type: ignore
        if mock_run_exception is None:
            assert state.result() is None
        else:
            with pytest.raises(mock_run_exception):
                state.result()

    @pytest.fixture(params=[None, MockError])
    def mock_run_exception(self, request) -> Any:  # type: ignore
        return request.param



from __future__ import annotations

import re

from nextline.utils import profile_func


def test_one():
    def func():
        re.compile("foo|bar")
        return 123

    result, ret = profile_func(func)
    assert result
    assert ret == 123



import pytest

from nextline.process.trace import TraceFromFactory

from .funcs import TraceSummary

from . import module_a


def test_one(
    target: TraceSummary,
    probe: TraceSummary,
    ref: TraceSummary,
):
    assert ref.call.module
    assert ref.return_.module
    assert ref.call == target.call
    assert not target.return_.module
    assert ref == probe


def f():
    module_a.func_a()


@pytest.fixture()
def func():
    yield f


@pytest.fixture(params=[set(), {module_a.__name__}])
def modules_to_skip(request):
    y = request.param
    yield y


@pytest.fixture()
def target_trace_func(mock_factory):
    y = TraceFromFactory(factory=mock_factory)
    yield y


@pytest.fixture()
def mock_factory(probe_trace_func):
    return lambda: probe_trace_func


@pytest.fixture()
def thread():
    yield False



from __future__ import annotations

from asyncio import Task
from functools import lru_cache
from threading import Thread
from weakref import WeakKeyDictionary
import fnmatch

from collections.abc import Set, MutableSet

from typing import TYPE_CHECKING, Optional, Callable
from types import FrameType

from .pdb.proxy import PdbInterfaceTraceFuncFactory
from ..utils import current_task_or_thread

if TYPE_CHECKING:
    from sys import _TraceFunc as TraceFunc
    from .run import Context


MODULES_TO_SKIP = {
    "threading",
    "queue",
    "importlib",
    "asyncio.*",
    "janus",
    "codec",
    "concurrent.futures.*",
    "selectors",
    "weakref",
    "_weakrefset",
    "socket",
    "logging",
    "os",
    "collections.*",
    "importlib.*",
    "pathlib",
    "typing",
    "posixpath",
    "fnmatch",
    "_pytest.*",
    "pluggy.*",
    "nextline.pdb.*",
    "nextline.utils.*",
    "nextline.queuedist",
    "nextlinegraphql.schema.bindables",
}


def Trace(context: Context) -> TraceFunc:
    """Create the main trace function"""

    modules_to_trace = context["modules_to_trace"]

    trace_func_factory = PdbInterfaceTraceFuncFactory(context=context)

    def create_trace_for_single_thread_or_task():
        """To be called in the thread or task to be traced"""
        trace_call_pdb = TraceFromFactory(factory=trace_func_factory)
        return TraceSelectFirstModule(
            trace=trace_call_pdb,
            modules_to_trace=modules_to_trace,
        )

    return TraceSkipModule(
        skip=MODULES_TO_SKIP,
        trace=TraceSkipLambda(
            trace=TraceAddFirstModule(
                modules_to_trace=modules_to_trace,
                trace=TraceDispatchThreadOrTask(
                    factory=create_trace_for_single_thread_or_task
                ),
            ),
        ),
    )


def TraceSkipModule(trace: TraceFunc, skip: Set[str]) -> TraceFunc:
    skip = frozenset(skip)

    @lru_cache
    def to_skip(module_name: str | None) -> bool:
        # NOTE: _is_matched_to_any() is slow
        return _is_matched_to_any(module_name, skip)

    def ret(frame: FrameType, event, arg) -> Optional[TraceFunc]:
        module_name = frame.f_globals.get("__name__")
        if to_skip(module_name):
            return None
        return trace(frame, event, arg)

    return ret


def TraceSkipLambda(trace: TraceFunc) -> TraceFunc:
    def ret(frame: FrameType, event, arg) -> Optional[TraceFunc]:
        func_name = frame.f_code.co_name
        if func_name == "<lambda>":
            return None
        return trace(frame, event, arg)

    return ret


def TraceAddFirstModule(
    trace: TraceFunc, modules_to_trace: MutableSet[str]
) -> TraceFunc:
    first = True

    def global_trace(frame: FrameType, event, arg) -> Optional[TraceFunc]:
        if not first:
            return trace(frame, event, arg)

        def create_local_trace() -> TraceFunc:
            next_trace: TraceFunc | None = trace

            def local_trace(frame, event, arg) -> Optional[TraceFunc]:
                nonlocal first, next_trace
                assert next_trace

                if module_name := frame.f_globals.get("__name__"):
                    first = False
                    modules_to_trace.add(module_name)
                    return trace(frame, event, arg)

                if next_trace := next_trace(frame, event, arg):
                    return local_trace
                return None

            return local_trace

        return create_local_trace()(frame, event, arg)

    return global_trace


def TraceDispatchThreadOrTask(factory: Callable[[], TraceFunc]) -> TraceFunc:
    """Create a trace that creates a new trace for each thread or asyncio task"""

    map: WeakKeyDictionary[Task | Thread, TraceFunc] = WeakKeyDictionary()

    def ret(frame, event, arg) -> Optional[TraceFunc]:
        key = current_task_or_thread()
        trace = map.get(key)
        if not trace:
            trace = factory()
            map[key] = trace
        return trace(frame, event, arg)

    return ret


def TraceSelectFirstModule(
    trace: TraceFunc,
    modules_to_trace: Set[str],
) -> TraceFunc:
    first = True

    def ret(frame: FrameType, event, arg) -> Optional[TraceFunc]:
        nonlocal first
        if first:
            module_name = frame.f_globals.get("__name__")
            if not _is_matched_to_any(module_name, modules_to_trace):
                return None
            first = False
        return trace(frame, event, arg)

    return ret


def TraceFromFactory(factory: Callable[[], TraceFunc]) -> TraceFunc:
    """Create a trace that creates a trace first time called

    Useful when desirable to defer the creation until actually called

    It is used to avoid creating instances of Pdb in threads or async tasks
    that are not traced.
    """
    trace: TraceFunc | None = None

    def global_trace(frame, event, arg) -> Optional[TraceFunc]:
        nonlocal trace
        trace = trace or factory()
        return trace(frame, event, arg)

    return global_trace


def _is_matched_to_any(word: str | None, patterns: Set[str]) -> bool:
    """Test if the word matches any of the patterns

    This function is based on Bdb.is_skipped_module():
    https://github.com/python/cpython/blob/v3.9.5/Lib/bdb.py#L191
    """
    if word is None:
        return False
    for pattern in patterns:
        if fnmatch.fnmatch(word, pattern):
            return True
    return False



from __future__ import annotations
from operator import itemgetter
from itertools import groupby


from nextline.utils import merge_aiters
from .aiterable import aiterable


async def test_one() -> None:
    obj = merge_aiters(aiterable(range(5)), aiterable([True, False] * 3))
    results = [i async for i in obj]
    actual = {
        k: list(map(itemgetter(1), v))
        for k, v in groupby(sorted(results, key=itemgetter(0)), itemgetter(0))
    }
    expected = {0: [0, 1, 2, 3, 4], 1: [True, False, True, False, True, False]}
    assert actual == expected



from __future__ import annotations
from concurrent.futures import ThreadPoolExecutor
from contextlib import contextmanager
from functools import partial

from queue import Queue

from ...types import PromptNo, TraceNo
from ...count import PromptNoCounter, TraceNoCounter
from .ci import pdb_command_interface
from .custom import CustomizedPdb
from .stream import StreamIn, StreamOut


from typing import (
    TYPE_CHECKING,
    Callable,
    Optional,
    Any,
    Tuple,
)
from types import FrameType

if TYPE_CHECKING:
    from sys import _TraceFunc as TraceFunc
    from ..run import Context


def PdbInterfaceTraceFuncFactory(context: Context) -> Callable[[], TraceFunc]:

    trace_no_counter = TraceNoCounter(1)
    prompt_no_counter = PromptNoCounter(1)

    def factory() -> TraceFunc:
        trace_no = trace_no_counter()
        pdbi = PdbInterface(
            trace_no=trace_no,
            context=context,
            prompt_no_counter=prompt_no_counter,
        )
        context["callback"].task_or_thread_start(trace_no, pdbi)
        return pdbi.trace

    return factory


class PdbInterface:
    """Instantiate Pdb and register its command loops"""

    def __init__(
        self,
        trace_no: TraceNo,
        context: Context,
        prompt_no_counter: Callable[[], PromptNo],
    ):
        self._trace_no = trace_no
        self._ci_map = context["pdb_ci_map"]
        self._callback = context["callback"]
        self._opened = False

        q_stdin: Queue[str] = Queue()
        self._q_stdout: Queue[str | None] = Queue()

        self._pdb = CustomizedPdb(
            pdbi=self,
            stdin=StreamIn(q_stdin),
            stdout=StreamOut(self._q_stdout),  # type: ignore
            nosigint=True,
            readrc=False,
        )

        self._trace_args: Optional[Tuple[FrameType, str, Any]] = None

        self._cmd_interface = partial(
            pdb_command_interface,
            trace_no=self._trace_no,
            prompt_no_counter=prompt_no_counter,
            queue_stdin=q_stdin,
            queue_stdout=self._q_stdout,
            callback=context["callback"],
            prompt=self._pdb.prompt,
        )

        self._executor = ThreadPoolExecutor(max_workers=1)

    def trace(self, frame, event, arg) -> Optional[TraceFunc]:
        """Call Pdb while storing trace args"""

        @contextmanager
        def capture(frame, event, arg):
            self._trace_args = (frame, event, arg)
            self._callback.trace_call_start(self._trace_no, self._trace_args)
            try:
                yield
            finally:
                self._trace_args = None
                self._callback.trace_call_end(self._trace_no)

        def create_local_trace() -> TraceFunc:
            next_trace: TraceFunc | None = self._pdb.trace_dispatch

            def local_trace(frame, event, arg) -> Optional[TraceFunc]:
                nonlocal next_trace
                assert next_trace
                with capture(frame, event, arg):
                    if next_trace := next_trace(frame, event, arg):
                        return local_trace
                    return None

            return local_trace

        return create_local_trace()(frame, event, arg)

    def entering_cmdloop(self) -> None:
        """To be called by the custom Pdb before _cmdloop()"""

        if not self._trace_args:
            raise RuntimeError("calling_trace() must be called first")

        wait_prompt, send_command = self._cmd_interface(
            trace_args=self._trace_args,
        )
        self._fut = self._executor.submit(wait_prompt)

        self._ci_map[self._trace_no] = send_command

    def exited_cmdloop(self) -> None:
        """To be called by the custom Pdb after _cmdloop()"""

        del self._ci_map[self._trace_no]
        self._q_stdout.put(None)  # end the thread
        self._fut.result()

    def close(self):
        self._executor.shutdown()



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Tuple

import pytest
from unittest.mock import Mock

from nextline.context import Context
from nextline.state import (
    State,
    Created,
    Initialized,
    Running,
    Finished,
    Closed,
    StateObsoleteError,
    StateMethodError,
)


class BaseTestState(ABC):
    """Test state classes of the state machine

    To be inherited by the test class for each state class.
    """

    state_class: Any = None

    @pytest.fixture
    def mock_run_exception(self) -> Any:
        return None

    @pytest.fixture
    def mock_run_result_exception(self, mock_run_exception) -> Tuple[Any, Any]:
        result = None
        return result, mock_run_exception

    @pytest.fixture
    def context(
        self,
        mock_run_result_exception: Tuple[Any, Any],
    ) -> Context:
        y = Mock(spec=Context)
        y.result, y.exception = mock_run_result_exception
        return y

    @pytest.fixture
    def created(self, context: Context) -> Created:
        return Created(context)

    @pytest.fixture
    async def initialized(self, created: Created) -> Initialized:
        return await created.initialize()

    @pytest.fixture
    async def running(self, initialized: Initialized) -> Running:
        return await initialized.run()

    @pytest.fixture()
    async def finished(self, running: Running) -> Finished:
        return await running.finish()

    @pytest.fixture()
    async def closed(self, finished: Finished) -> Closed:
        return await finished.close()

    @abstractmethod
    def state(self, _):
        """Yield an instance of the class being tested

        To be overridden as a pytest fixture.
        """
        pass

    def test_state(self, state: State, context: Mock):
        del context
        assert self.state_class is not None
        assert isinstance(state, self.state_class)
        assert "obsolete" not in repr(state)

    async def assert_obsolete(self, state: State):
        assert "obsolete" in repr(state)

        with pytest.raises(StateObsoleteError):
            await state.run()

        with pytest.raises(StateObsoleteError):
            await state.finish()

        with pytest.raises(StateObsoleteError):
            await state.reset()

        with pytest.raises(StateObsoleteError):
            await state.close()

    async def test_initialize(self, state: State):
        with pytest.raises(StateMethodError):
            await state.initialize()

    async def test_run(self, state: State):
        with pytest.raises(StateMethodError):
            await state.run()

    async def test_finish(self, state: State):
        with pytest.raises(StateMethodError):
            await state.finish()

    async def test_reset(self, state: State):
        with pytest.raises(StateMethodError):
            await state.reset()

    def test_exception(self, state: State):
        with pytest.raises(StateMethodError):
            state.exception()

    def test_result(self, state: State):
        with pytest.raises(StateMethodError):
            state.result()



from __future__ import annotations

import os
from asyncio import Task
import threading
from threading import Thread
from weakref import WeakKeyDictionary
import datetime
import dataclasses

from typing import TYPE_CHECKING, Any, Callable, Optional, Tuple, Set
from typing import Dict, MutableMapping  # noqa F401
from typing_extensions import TypeAlias
from types import FrameType

from ..types import RunNo, TraceNo, PromptNo, TraceInfo, PromptInfo, StdoutInfo
from ..utils import (
    ThreadTaskDoneCallback,
    ThreadTaskIdComposer,
    current_task_or_thread,
)
from .io import peek_stdout_by_task_and_thread

if TYPE_CHECKING:
    from .pdb.proxy import PdbInterface
    from .run import QueueRegistry

TraceNoMap: TypeAlias = "MutableMapping[Task | Thread, TraceNo]"
TraceInfoMap: TypeAlias = "Dict[TraceNo, TraceInfo]"
PromptInfoMap: TypeAlias = "Dict[Tuple[TraceNo, PromptNo], PromptInfo]"
PdbIMap: TypeAlias = "Dict[TraceNo, PdbInterface]"


class Callback:
    def __init__(
        self,
        run_no: RunNo,
        registrar: RegistrarProxy,
        modules_to_trace: Set[str],
    ):
        self._run_no = run_no
        self._registrar = registrar
        self._modules_to_trace = modules_to_trace
        self._trace_nos: Tuple[TraceNo, ...] = ()
        self._trace_no_map: TraceNoMap = WeakKeyDictionary()
        self._trace_id_factory = ThreadTaskIdComposer()
        self._trace_info_map: TraceInfoMap = {}
        self._thread_task_done_callback = ThreadTaskDoneCallback(
            done=self.task_or_thread_end
        )
        self._tasks_and_threads: Set[Task | Thread] = set()
        self._prompt_info_map: PromptInfoMap = {}
        self._pdbi_map: PdbIMap = {}
        self._to_canonic = ToCanonic()
        self._entering_thread: Optional[Thread] = None
        self._last_prompt_frame_map: Dict[TraceNo, FrameType] = {}
        self._current_trace_call_map: Dict[
            TraceNo, Tuple[FrameType, str, Any]
        ] = {}

    def task_or_thread_start(
        self, trace_no: TraceNo, pdbi: PdbInterface
    ) -> None:
        self.trace_start(trace_no, pdbi)

        task_or_thread = current_task_or_thread()
        self._trace_no_map[task_or_thread] = trace_no

        if task_or_thread is not self._entering_thread:
            self._thread_task_done_callback.register(task_or_thread)

        self._tasks_and_threads.add(task_or_thread)

    def task_or_thread_end(self, task_or_thread: Task | Thread):
        trace_no = self._trace_no_map[task_or_thread]
        self.trace_end(trace_no)

    def trace_start(self, trace_no: TraceNo, pdbi: PdbInterface):
        self._pdbi_map[trace_no] = pdbi

        # TODO: Putting a prompt info for now because otherwise tests get stuck
        # sometimes for an unknown reason. Need to investigate
        prompt_info = PromptInfo(
            run_no=self._run_no,
            trace_no=trace_no,
            prompt_no=PromptNo(-1),
            open=False,
        )
        self._registrar.put_prompt_info_for_trace(trace_no, prompt_info)

        self._trace_nos = self._trace_nos + (trace_no,)
        self._registrar.put_trace_nos(self._trace_nos)

        thread_task_id = self._trace_id_factory()

        trace_info = TraceInfo(
            run_no=self._run_no,
            trace_no=trace_no,
            thread_no=thread_task_id.thread_no,
            task_no=thread_task_id.task_no,
            state="running",
            started_at=datetime.datetime.now(),
        )
        self._trace_info_map[trace_no] = trace_info
        self._registrar.put_trace_info(trace_info)

    def trace_end(self, trace_no: TraceNo):
        self._registrar.end_prompt_info_for_trace(trace_no)
        self._pdbi_map.pop(trace_no).close()

        nosl = list(self._trace_nos)
        nosl.remove(trace_no)
        self._trace_nos = tuple(nosl)
        self._registrar.put_trace_nos(self._trace_nos)

        trace_info = self._trace_info_map[trace_no]

        trace_info = dataclasses.replace(
            trace_info,
            state="finished",
            ended_at=datetime.datetime.now(),
        )

        self._registrar.put_trace_info(trace_info)

    def trace_call_start(
        self,
        trace_no: TraceNo,
        trace_args: Tuple[FrameType, str, Any],
    ):
        self._current_trace_call_map[trace_no] = trace_args

    def trace_call_end(self, trace_no: TraceNo):
        frame, event, _ = self._current_trace_call_map[trace_no]
        file_name = self._to_canonic(frame.f_code.co_filename)
        line_no = frame.f_lineno

        if frame is not self._last_prompt_frame_map.get(trace_no):
            return

        prompt_info = PromptInfo(
            run_no=self._run_no,
            trace_no=trace_no,
            prompt_no=PromptNo(-1),
            open=False,
            event=event,
            file_name=file_name,
            line_no=line_no,
            trace_call_end=True,
        )
        self._registrar.put_prompt_info(prompt_info)
        self._registrar.put_prompt_info_for_trace(trace_no, prompt_info)

    def prompt_start(
        self,
        trace_no: TraceNo,
        prompt_no: PromptNo,
        trace_args: Tuple[FrameType, str, Any],
        out: str,
    ) -> None:
        frame, event, _ = trace_args
        file_name = self._to_canonic(frame.f_code.co_filename)
        line_no = frame.f_lineno
        if module_name := frame.f_globals.get("__name__"):
            self._modules_to_trace.add(module_name)
        prompt_info = PromptInfo(
            run_no=self._run_no,
            trace_no=trace_no,
            prompt_no=prompt_no,
            open=True,
            event=event,
            file_name=file_name,
            line_no=line_no,
            stdout=out,
            started_at=datetime.datetime.now(),
        )
        self._prompt_info_map[(trace_no, prompt_no)] = prompt_info
        self._registrar.put_prompt_info(prompt_info)
        self._registrar.put_prompt_info_for_trace(trace_no, prompt_info)
        self._last_prompt_frame_map[trace_no] = frame

    def prompt_end(
        self, trace_no: TraceNo, prompt_no: PromptNo, command: str
    ) -> None:
        prompt_info = self._prompt_info_map.pop((trace_no, prompt_no))
        prompt_info = dataclasses.replace(
            prompt_info,
            open=False,
            command=command,
            ended_at=datetime.datetime.now(),
        )
        self._registrar.put_prompt_info(prompt_info)
        self._registrar.put_prompt_info_for_trace(trace_no, prompt_info)

    def stdout(self, task_or_thread: Task | Thread, line: str):
        trace_no = self._trace_no_map[task_or_thread]
        stdout_info = StdoutInfo(
            run_no=self._run_no,
            trace_no=trace_no,
            text=line,
            written_at=datetime.datetime.now(),
        )
        self._registrar.put_stdout_info(stdout_info)

    def close(self) -> None:
        self._thread_task_done_callback.close()

    def __enter__(self):
        self._peek_stdout = peek_stdout_by_task_and_thread(
            to_peek=self._tasks_and_threads, callback=self.stdout
        )
        self._entering_thread = threading.current_thread()
        self._peek_stdout.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self._peek_stdout.__exit__(exc_type, exc_value, traceback)
        self.close()
        if self._entering_thread:
            if trace_no := self._trace_no_map.get(self._entering_thread):
                self.trace_end(trace_no)


class RegistrarProxy:
    def __init__(self, queue: QueueRegistry):
        self._queue = queue

    def put_trace_nos(self, trace_nos: Tuple[TraceNo, ...]) -> None:
        self._queue.put(("trace_nos", trace_nos, False))

    def put_trace_info(self, trace_info: TraceInfo) -> None:
        self._queue.put(("trace_info", trace_info, False))

    def put_prompt_info(self, prompt_info: PromptInfo) -> None:
        self._queue.put(("prompt_info", prompt_info, False))

    def put_prompt_info_for_trace(
        self, trace_no: TraceNo, prompt_info: PromptInfo
    ) -> None:
        key = f"prompt_info_{trace_no}"
        self._queue.put((key, prompt_info, False))

    def end_prompt_info_for_trace(self, trace_no: TraceNo) -> None:
        key = f"prompt_info_{trace_no}"
        self._queue.put((key, None, True))

    def put_stdout_info(self, stdout_info: StdoutInfo) -> None:
        self._queue.put(("stdout", stdout_info, False))


def ToCanonic() -> Callable[[str], str]:
    # Based on Bdb.canonic()
    # https://github.com/python/cpython/blob/v3.10.5/Lib/bdb.py#L39-L54

    cache: Dict[str, str] = {}

    def to_canonic(filename: str) -> str:
        if filename == "<" + filename[1:-1] + ">":
            return filename
        canonic = cache.get(filename)
        if not canonic:
            canonic = os.path.abspath(filename)
            canonic = os.path.normcase(canonic)
            cache[filename] = canonic
        return canonic

    return to_canonic



from __future__ import annotations

import pytest
from unittest.mock import Mock, call, sentinel

from nextline.state import State, Initialized, Running, Closed

from .base import BaseTestState


class TestInitialized(BaseTestState):

    state_class = Initialized

    @pytest.fixture()
    def state(self, initialized: State) -> State:
        return initialized

    def test_state(self, state: State, context: Mock):
        super().test_state(state, context)
        assert [call.initialize(state)] == context.mock_calls

    async def test_run(self, state: State):
        running = await state.run()
        assert isinstance(running, Running)
        await self.assert_obsolete(state)

    async def test_reset(self, state: State, context: Mock):  # type: ignore
        reset = await state.reset(sentinel.args)
        assert isinstance(reset, Initialized)
        assert reset is not state
        await self.assert_obsolete(state)
        assert [call(sentinel.args)] == context.reset.call_args_list

    async def test_close(self, state: State):
        closed = await state.close()
        assert isinstance(closed, Closed)
        await self.assert_obsolete(state)



from __future__ import annotations

from itertools import groupby
from operator import itemgetter
import random
import string
import time
from threading import Thread
from typing import Iterator, Optional, TextIO


from nextline.process.io import peek_stdout_by_task_and_thread


from unittest.mock import Mock


def print_lines(lines: Iterator[str], file: Optional[TextIO] = None):
    for line in lines:
        time.sleep(0.0001)
        print(line, file=file)


def test_one(capsys):
    config = (
        (tuple(random_strings()), True),
        (tuple(random_strings()), True),
        (tuple(random_strings()), False),
    )
    thread_added = tuple(
        (Thread(target=print_lines, args=(c[0],)),) + c for c in config
    )
    threads = tuple(c[0] for c in thread_added)
    to_return = [c[0] for c in thread_added if c[2]]
    callback = Mock()
    with peek_stdout_by_task_and_thread(to_peek=to_return, callback=callback):
        for t in threads:
            t.start()
        for t in threads:
            t.join()
    results = {
        k: "".join(v[1] for v in v)
        for k, v in groupby(
            sorted(
                (c.args for c in callback.call_args_list),
                key=lambda args: args[0].name,
            ),
            itemgetter(0),
        )
    }
    expected = {c[0]: "\n".join(c[1]) + "\n" for c in thread_added if c[2]}
    assert results == expected


def test_random_strings():
    lines = random_strings()
    assert all(isinstance(line, str) for line in lines)


def random_strings():
    n_lines_range = (100, 200)
    # n_lines_range = (3, 10)
    n_letters_per_line_range = (0, 300)
    # n_letters_per_line_range = (0, 5)
    # letters = string.printable
    letters = string.ascii_letters
    r = (
        "".join(
            random.choice(letters)
            for _ in range(random.randint(*n_letters_per_line_range))
        )
        for _ in range(random.randint(*n_lines_range))
    )
    return r



import asyncio
from random import random, randint

import pytest

from typing import List, Set

from nextline.utils import agen_with_wait


async def test_one():
    async def agen():
        for i in range(3):
            yield i
            await asyncio.sleep(0.001)

    async def afunc():
        delay = random() * 0.001
        await asyncio.sleep(delay)

    all: Set[asyncio.Task] = set()
    done: List[asyncio.Task] = []

    obj = agen_with_wait(agen())
    async for _ in obj:
        tasks = {asyncio.create_task(afunc()) for _ in range(randint(0, 5))}
        all |= tasks
        done_, pending = await obj.asend(tasks)
        done.extend(done_)

    await asyncio.gather(*pending)
    assert len(all) == len(done) + len(pending)
    assert all == set(done) | set(pending)


async def test_raise():
    async def agen():
        yield 0
        await asyncio.sleep(0.1)
        assert False  # The line shouldn't be reached

    async def afunc():
        await asyncio.sleep(0)
        raise Exception("foo", "bar")

    obj = agen_with_wait(agen())
    with pytest.raises(Exception) as exc:
        async for _ in obj:
            tasks = {asyncio.create_task(afunc())}
            _, pending = await obj.asend(tasks)

    assert ("foo", "bar") == exc.value.args


async def test_without_send():
    async def agen():
        for i in range(3):
            yield i
            await asyncio.sleep(0)

    items = []
    async for i in agen_with_wait(agen()):
        # print(i)
        items.append(i)

    assert [0, 1, 2] == items



from __future__ import annotations

import dataclasses

from typing import Iterable, TypeVar


_T = TypeVar("_T")


def replace_with_bool(obj: _T, fields: Iterable[str]) -> _T:
    changes = {f: not not getattr(obj, f) for f in fields}
    return dataclasses.replace(obj, **changes)



import time
import random

from threading import Thread
import asyncio

import pytest


from nextline.utils import ThreadTaskDoneCallback, current_task_or_thread


def target(obj: ThreadTaskDoneCallback):
    """To run in a thread or task"""
    assert current_task_or_thread() == obj.register()
    delay = random.random() * 0.01
    time.sleep(delay)


async def atarget(obj: ThreadTaskDoneCallback):
    target(obj)


class Done:
    """A callback function"""

    def __init__(self):
        self.args = set()

    def __call__(self, arg):
        self.args.add(arg)


@pytest.fixture()
def done():
    """A callback function"""
    yield Done()


def test_thread(done: Done):
    obj = ThreadTaskDoneCallback(done=done)
    t = Thread(target=target, args=(obj,))
    t.start()
    time.sleep(0.005)
    obj.close()
    assert {t} == done.args
    t.join()


async def test_task(done: Done):
    obj = ThreadTaskDoneCallback(done=done)
    t = asyncio.create_task(atarget(obj))
    await t
    await obj.aclose()
    assert {t} == done.args


def test_with_thread(done: Done):
    with ThreadTaskDoneCallback(done=done) as obj:
        t = Thread(target=target, args=(obj,))
        t.start()
        time.sleep(0.005)
    assert {t} == done.args
    t.join()


async def test_with_task(done: Done):
    async with ThreadTaskDoneCallback(done=done) as obj:
        t = asyncio.create_task(atarget(obj))
        await t
    assert {t} == done.args


def test_done_none_thread():
    with ThreadTaskDoneCallback() as obj:
        t = Thread(target=target, args=(obj,))
        t.start()
        time.sleep(0.005)
    assert not t.is_alive()
    t.join()


async def test_done_none_task():
    async with ThreadTaskDoneCallback() as obj:
        t = asyncio.create_task(atarget(obj))
        await asyncio.sleep(0)  # let the task be registered
    assert t.done()  # finished after exited
    await t



from setuptools import setup, find_packages
import versioneer

from pathlib import Path

here = Path(__file__).resolve().parent
long_description = here.joinpath("README.md").read_text()

setup(
    name="nextline",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="A Python library for controlling Python scripts execution",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Simons Observatory",
    author_email="so_software@simonsobservatory.org",
    url="https://github.com/simonsobs/nextline",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    packages=find_packages(exclude=["docs", "tests"]),
    include_package_data=True,
    install_requires=[
        "janus>=0.6",
        "tblib>=1.7",
    ],
    extras_require={
        "tests": [
            "pytest>=7.0",
            "pytest-cov>=3.0",
            "pytest-asyncio>=0.18",
            "pytest-timeout>=2.1",
            "snapshottest>=0.6",
        ],
        "dev": [
            "tox",
            "twine",
            "flake8",
            "black",
            "mypy",
        ],
    },
)



from __future__ import annotations

from types import FrameType
from queue import Queue
from typing import Callable, Tuple, Any

from ..callback import Callback
from ...types import TraceNo, PromptNo


def pdb_command_interface(
    trace_args: Tuple[FrameType, str, Any],
    trace_no: TraceNo,
    prompt_no_counter: Callable[[], PromptNo],
    queue_stdin: Queue[str],
    queue_stdout: Queue[str | None],
    callback: Callback,
    prompt="(Pdb) ",
) -> Tuple[Callable[[], None], Callable[[str, PromptNo], None]]:

    _prompt_no: PromptNo

    def wait_prompt() -> None:
        """receive stdout from pdb

        To be run in a thread during pdb._cmdloop()
        """
        nonlocal _prompt_no
        while out := _read_until_prompt(queue_stdout, prompt):
            _prompt_no = prompt_no_counter()
            # print(_prompt_no)
            callback.prompt_start(
                trace_no=trace_no,
                prompt_no=_prompt_no,
                trace_args=trace_args,
                out=out,
            )

    def send_command(command: str, prompt_no: PromptNo) -> None:
        """send a command to pdb"""
        callback.prompt_end(
            trace_no=trace_no, prompt_no=prompt_no, command=command
        )
        queue_stdin.put(command)

    return wait_prompt, send_command


def _read_until_prompt(queue: Queue[str | None], prompt: str) -> str | None:
    """read the queue up to the prompt"""
    out = ""
    while (m := queue.get()) is not None:
        out += m
        if prompt == m:
            break
    else:
        return None

    return out



import asyncio
from string import ascii_lowercase, ascii_uppercase
import time

import pytest

from nextline.utils import PubSub


@pytest.fixture
async def obj():
    async with PubSub() as y:
        yield y


async def test_end(obj: PubSub[str, str]):
    key = "A"

    async def subscribe():
        return tuple([y async for y in obj.subscribe(key)])

    async def put():
        await asyncio.sleep(0.001)
        await obj.end(key)

    result, _ = await asyncio.gather(subscribe(), put())
    assert result == ()


async def test_end_without_subscription(obj: PubSub[str, str]):
    key = "A"
    await obj.end(key)


async def test_close(obj: PubSub[str, str]):
    key = "A"
    n_items = 5
    items = tuple(ascii_lowercase[:n_items])

    async def subscribe():
        return tuple([y async for y in obj.subscribe(key)])

    async def put():
        await asyncio.sleep(0.001)
        for item in items:
            await obj.publish(key, item)
        assert obj.latest(key) == items[-1]
        await obj.close()  # ends the subscription
        with pytest.raises(LookupError):
            obj.latest(key)

    result, _ = await asyncio.gather(subscribe(), put())
    assert result == items


@pytest.mark.parametrize("last", [True, False])
async def test_last(obj: PubSub[str, str], last: bool):
    key = "A"
    n_pre_items = 3
    n_items = 5
    pre_items = tuple(reversed(ascii_lowercase[-n_pre_items:]))
    items = tuple(ascii_lowercase[:n_items])
    expected = pre_items[-1:] + items if last else items

    for item in pre_items:
        await obj.publish(key, item)

    await asyncio.sleep(0.001)

    async def subscribe():
        return tuple([y async for y in obj.subscribe(key, last=last)])

    async def put():
        await asyncio.sleep(0.001)
        for item in items:
            await obj.publish(key, item)
        await obj.end(key)

    result, _ = await asyncio.gather(subscribe(), put())
    assert result == expected


@pytest.mark.parametrize("n_keys", [0, 1, 3])
@pytest.mark.parametrize("n_items", [0, 1, 2, 30])
@pytest.mark.parametrize("n_subscribers", [0, 1, 2, 50])
async def test_matrix(
    obj: PubSub[str, str],
    n_keys: int,
    n_items: int,
    n_subscribers: int,
):
    keys = ascii_uppercase[:n_keys]
    items = {k: tuple(f"{k}-{i+1}" for i in range(n_items)) for k in keys}

    async def subscribe(key):
        return tuple([y async for y in obj.subscribe(key)])

    async def put(key):
        time.sleep(0.01)
        for item in items[key]:
            await obj.publish(key, item)
        await obj.end(key)

    results = await asyncio.gather(
        *(subscribe(k) for k in keys for _ in range(n_subscribers)),
        *(put(k) for k in keys),
    )
    actual = results[: n_subscribers * n_keys]
    expected = [items[k] for k in keys for _ in range(n_subscribers)]
    assert actual == expected



from __future__ import annotations

import asyncio
from functools import partial
from typing import Any, Callable, TypeVar

T = TypeVar("T")


def ToLoop() -> Callable[[Callable[[Any], T]], T]:
    """Create a function that calls a function in the same asyncio event loop

    A running asyncio event loop needs to exist in the thread in which this
    function is called. The returned function can be called in any thread. A
    function given to the returned function will be called in the event loop in
    the thread in which this function is called.
    """
    loop = asyncio.get_running_loop()

    def in_the_same_running_loop():
        try:
            loop_ = asyncio.get_running_loop()
        except RuntimeError:
            return False
        return loop is loop_

    def to_loop(func: Callable[[Any], T], /, *args, **kwargs) -> T:
        func_ = partial(func, *args, **kwargs)

        if in_the_same_running_loop():
            return func_()

        if loop.is_closed():
            raise RuntimeError(f"The loop is closed: {loop}")

        # In another thread

        async def afunc():
            return func_()

        fut = asyncio.run_coroutine_threadsafe(afunc(), loop)
        return fut.result()

    return to_loop



import pytest
from unittest.mock import Mock

from typing import Set

from nextline.process.trace import TraceSkipModule

from .funcs import TraceSummary

from . import module_a


def test_one(
    target: TraceSummary,
    probe: TraceSummary,
    ref: TraceSummary,
    modules_to_skip: Set[str],
):
    assert ref.call.module
    assert ref.return_.module
    assert ref.call.module == target.call.module
    assert not target.return_.module
    assert set(ref.call.module) - modules_to_skip == set(probe.call.module)
    assert set(ref.call.module) - modules_to_skip == set(probe.return_.module)


def f():
    module_a.func_a()


@pytest.fixture()
def func():
    yield f


@pytest.fixture(params=[set(), {module_a.__name__}])
def modules_to_skip(request):
    y = request.param
    yield y


@pytest.fixture()
def target_trace_func(probe_trace_func: Mock, modules_to_skip: Set[str]):
    y = TraceSkipModule(
        trace=probe_trace_func,
        skip=modules_to_skip,
    )
    yield y


@pytest.fixture()
def thread():
    yield False



from __future__ import annotations
import sys

import pytest
from unittest.mock import Mock, call

from nextline.utils import peek_stdout, peek_stderr


def test_stdout(capsys):
    callback = Mock()

    with peek_stdout(callback):
        print("foo")

    print("bar")

    # with capsys.disabled():
    #     print(callback.call_args_list)

    assert [call("foo"), call("\n")] == callback.call_args_list

    captured = capsys.readouterr()
    assert "foo\nbar\n" == captured.out


def test_stderr(capsys):
    callback = Mock()

    with peek_stderr(callback):
        print("foo", file=sys.stderr)

    print("bar", file=sys.stderr)

    # with capsys.disabled():
    #     print(callback.call_args_list)

    assert [call("foo"), call("\n")] == callback.call_args_list

    captured = capsys.readouterr()
    assert "foo\nbar\n" == captured.err


def test_stdout_target(capsys):
    callback = Mock()

    with peek_stdout(callback) as t:
        t("foo")

    assert [call("foo")] == callback.call_args_list

    captured = capsys.readouterr()
    assert "foo" == captured.out


def test_stderr_target(capsys):
    callback = Mock()

    with peek_stderr(callback) as t:
        t("foo")

    assert [call("foo")] == callback.call_args_list

    captured = capsys.readouterr()
    assert "foo" == captured.err


def test_raise(capsys):
    callback = Mock(side_effect=MockError)

    with peek_stdout(callback):
        with pytest.raises(MockError):
            print("foo")

    print("bar")

    assert [call("foo")] == callback.call_args_list

    captured = capsys.readouterr()
    assert "bar\n" == captured.out


class MockError(BaseException):
    pass



from __future__ import annotations

import linecache
from logging import getLogger

from typing import Any, AsyncIterator, Optional, Tuple

from .utils import merge_aiters
from .context import Context
from .state import Machine
from .types import PromptNo, TraceNo, StdoutInfo

from .types import RunInfo, TraceInfo, PromptInfo


class Nextline:
    """Nextline allows line-by-line execution of concurrent Python scripts

    Nextline supports concurrency with threading and asyncio. It uses multiple
    instances of Pdb, one for each thread and async task.

    Note
    ----
    The running asyncio event loop must exists when Nextline is instantiated.

    Parameters
    ----------
    statement : str
        A Python code as a string

    """

    def __init__(self, statement: str, run_no_start_from: int = 1):

        logger = getLogger(__name__)
        logger.debug(f"statement starts with {statement[:25]!r}")
        logger.debug(f"The next run number will be {run_no_start_from}")

        self._context = Context(
            run_no_start_from=run_no_start_from,
            statement=statement,
        )
        self._registry = self._context.registry
        self._q_commands = self._context.q_commands

        self._closed = False

    async def start(self) -> None:
        await self._context.start()
        self._machine = Machine(self._context)
        await self._machine.initialize()

    def __repr__(self):
        # e.g., "<Nextline 'running'>"
        return f"<{self.__class__.__name__} {self.state!r}>"

    async def close(self) -> None:
        """End gracefully"""
        if self._closed:
            return
        self._closed = True
        await self._machine.close()
        await self._context.shutdown()
        await self._registry.close()

    async def __aenter__(self):
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        del exc_type, exc_value, traceback
        await self.close()

    async def run(self) -> None:
        """Execute the script and wait until it exits"""
        await self._machine.run()
        await self._machine.finish()

    def send_pdb_command(
        self, command: str, prompt_no: int, trace_no: int
    ) -> None:
        self._q_commands.put((command, PromptNo(prompt_no), TraceNo(trace_no)))

    def interrupt(self) -> None:
        self._machine.interrupt()

    def terminate(self) -> None:
        self._machine.terminate()

    def kill(self) -> None:
        self._machine.kill()

    def exception(self) -> Optional[BaseException]:
        """Uncaught exception from the last run"""
        return self._machine.exception()

    def result(self) -> Any:
        """Return value of the last run. always None"""
        return self._machine.result()

    async def reset(
        self,
        statement: Optional[str] = None,
        run_no_start_from: Optional[int] = None,
    ) -> None:
        """Prepare for the next run"""
        await self._machine.reset(
            statement=statement,
            run_no_start_from=run_no_start_from,
        )

    @property
    def statement(self) -> str:
        """The script"""
        return self.get("statement")

    @property
    def state(self) -> str:
        """The current condition of the script execution.

        The possible values are "initialized", "running", "finished",
        "closed"
        """
        return self._machine.state_name

    def subscribe_state(self) -> AsyncIterator[str]:
        return self.subscribe("state_name")

    @property
    def run_no(self):
        """The current run number"""
        return self.get("run_no")

    def subscribe_run_no(self) -> AsyncIterator[int]:
        return self.subscribe("run_no")

    def subscribe_trace_ids(self) -> AsyncIterator[Tuple[int]]:
        return self.subscribe("trace_nos")

    def get_source(self, file_name=None):
        if not file_name or file_name == self._registry.latest(
            "script_file_name"
        ):
            return self.get("statement").split("\n")
        return [e.rstrip() for e in linecache.getlines(file_name)]

    def get_source_line(self, line_no, file_name=None):
        """
        based on linecache.getline()
        https://github.com/python/cpython/blob/v3.9.5/Lib/linecache.py#L26
        """
        lines = self.get_source(file_name)
        if 1 <= line_no <= len(lines):
            return lines[line_no - 1]
        return ""

    def subscribe_run_info(self) -> AsyncIterator[RunInfo]:
        return self.subscribe("run_info")

    def subscribe_trace_info(self) -> AsyncIterator[TraceInfo]:
        return self.subscribe("trace_info")

    def subscribe_prompt_info(self) -> AsyncIterator[PromptInfo]:
        return self.subscribe("prompt_info")

    def subscribe_prompt_info_for(
        self, trace_no: int
    ) -> AsyncIterator[PromptInfo]:
        return self.subscribe(f"prompt_info_{trace_no}")

        # Alternative implementation under development
        # return self._subscribe_prompt_info_for(trace_no)

    async def _subscribe_prompt_info_for(
        self, trace_no: int
    ) -> AsyncIterator[PromptInfo]:

        merged = merge_aiters(
            self.subscribe_prompt_info(),
            self.subscribe_trace_info(),
        )
        async for _, info in merged:
            if not info.trace_no == trace_no:  # type: ignore
                continue
            if isinstance(info, TraceInfo):
                if info.trace_no == trace_no:
                    if info.state == "finished":
                        break
                continue
            assert isinstance(info, PromptInfo)
            yield info

    def get(self, key) -> Any:
        return self._registry.latest(key)

    def subscribe(
        self, key, last: Optional[bool] = True
    ) -> AsyncIterator[Any]:
        return self._registry.subscribe(key, last=last)

    def subscribe_stdout(self) -> AsyncIterator[StdoutInfo]:
        return self.subscribe("stdout", last=False)



from __future__ import annotations

from dataclasses import dataclass, InitVar, field
from functools import partial
import json
import multiprocessing as mp
from concurrent.futures import ProcessPoolExecutor
import traceback
from typing import TYPE_CHECKING, Callable, Coroutine, Optional, Any
from typing_extensions import ParamSpec

from .utils import (
    PubSub,
    MultiprocessingLogging,
    run_in_process,
    RunInProcess,
)
from .process import run
from .process.run import QueueRegistry, RunArg, QueueCommands
from .registrar import Registrar
from .types import RunNo
from .count import RunNoCounter

if TYPE_CHECKING:
    from .state import State

SCRIPT_FILE_NAME = "<string>"

_P = ParamSpec("_P")


def _initializer(
    init_logging: Callable[[], Any],
    initializer: Callable[_P, Any],
    *initargs: _P.args,
    **initkwargs: _P.kwargs,
) -> None:
    init_logging()
    if initializer is not None:
        initializer(*initargs, **initkwargs)


@dataclass
class Context:
    run_no_start_from: InitVar[int]
    statement: str
    filename: str = SCRIPT_FILE_NAME
    func: Callable = run.run
    state: Optional[State] = None
    future: Optional[RunInProcess] = None
    result: Optional[Any] = None
    exception: Optional[BaseException] = None
    registry: PubSub[Any, Any] = field(init=False)
    q_commands: QueueCommands = field(init=False)
    registrar: Registrar = field(init=False)
    run_no: RunNo = field(init=False)
    run_no_count: Callable[[], RunNo] = field(init=False)
    mp_logging: MultiprocessingLogging = field(init=False)
    runner: Callable[
        ...,
        Coroutine[Any, Any, RunInProcess],
    ] = field(init=False)

    def __post_init__(self, run_no_start_from: int):
        self.registry = PubSub[Any, Any]()
        mp_context = mp.get_context("spawn")
        self.q_commands: QueueCommands = mp_context.Queue()
        q_registry: QueueRegistry = mp_context.Queue()
        self.mp_logging = MultiprocessingLogging(context=mp_context)
        executor_factory = partial(
            ProcessPoolExecutor,
            max_workers=1,
            mp_context=mp_context,
            initializer=partial(
                _initializer,
                self.mp_logging.init,
                run.set_queues,
            ),
            initargs=(self.q_commands, q_registry),
        )
        self.runner = partial(run_in_process, executor_factory)  # type: ignore
        self.registrar = Registrar(self.registry, q_registry)
        self.run_no = RunNo(run_no_start_from - 1)
        self.run_no_count = RunNoCounter(run_no_start_from)

    async def start(self):
        await self.registrar.script_change(
            script=self.statement, filename=self.filename
        )

    async def shutdown(self):
        await self.registrar.close()
        await self.mp_logging.close()

    async def initialize(self, state: State):
        self.run_no = self.run_no_count()
        self.result = None
        self.exception = None
        await self.registrar.state_initialized(self.run_no)
        await self.registrar.state_change(state)
        self.state = state

    async def reset(
        self,
        statement: Optional[str] = None,
        run_no_start_from: Optional[int] = None,
    ):
        if statement:
            self.statement = statement
            await self.registrar.script_change(
                script=statement, filename=self.filename
            )
        if run_no_start_from is not None:
            self.run_no_count = RunNoCounter(run_no_start_from)

    async def run(self, state: State) -> RunInProcess:
        self.future = await self.runner(
            self.func,
            RunArg(
                run_no=self.run_no,
                statement=self.statement,
                filename=self.filename,
            ),
        )
        await self.registrar.run_start(self.run_no)
        await self.registrar.state_change(state)
        self.state = state
        return self.future

    def interrupt(self) -> None:
        if self.future:
            self.future.interrupt()

    def terminate(self) -> None:
        if self.future:
            self.future.terminate()

    def kill(self) -> None:
        if self.future:
            self.future.kill()

    async def finish(self, state: State) -> None:
        assert self.future
        try:
            self.result, self.exception = await self.future
        except TypeError:
            # The process was terminated.
            pass

        if self.exception:
            ret = None
            fmt_exc = "".join(
                traceback.format_exception(
                    type(self.exception),
                    self.exception,
                    self.exception.__traceback__,
                )
            )
        else:
            ret = json.dumps(self.result)
            fmt_exc = None

        await self.registrar.run_end(result=ret, exception=fmt_exc)
        await self.registrar.state_change(state)
        self.state = state

    async def close(self, state: State):
        await self.registrar.state_change(state)
        self.state = state



"""Test the class Machine

"""

import asyncio
from unittest.mock import Mock
import pytest

from nextline.context import Context
from nextline.state import Machine


SOURCE = """
import time
time.sleep(0.001)
""".strip()

SOURCE_TWO = """
x = 2
""".strip()


async def test_init(context: Context):
    async with Machine(context) as obj:
        assert "initialized" == obj.state_name


async def test_repr(context: Context):
    async with Machine(context) as obj:
        repr(obj)


async def test_state_name_unknown(monkeypatch, context: Context):
    async with Machine(context) as obj:
        with monkeypatch.context() as m:
            m.setattr(obj, "_state", None)
            assert "unknown" == obj.state_name
            del obj._state
            assert "unknown" == obj.state_name


async def test_transitions(context: Context):
    async with Machine(context) as obj:
        await asyncio.sleep(0)
        assert "initialized" == obj.state_name
        await obj.run()
        assert "running" == obj.state_name
        await obj.finish()
        assert "finished" == obj.state_name
        obj.result()
        obj.exception()
        await obj.reset()
        assert "initialized" == obj.state_name
    assert "closed" == obj.state_name


async def test_reset_with_statement(context: Context):
    async with Machine(context) as obj:
        await asyncio.sleep(0)
        await obj.run()
        await obj.finish()
        await obj.reset(statement=SOURCE_TWO)
        await obj.run()
        await obj.finish()


@pytest.fixture
def context() -> Mock:
    y = Mock(spec=Context)
    y.exception = None
    return y



__all__ = [
    "PubSubItem",
    "PubSub",
    "ThreadTaskIdComposer",
    "ThreadSafeAsyncioEvent",
    "ToLoop",
    "MultiprocessingLogging",
    "peek_stdout",
    "peek_stderr",
    "peek_textio",
    "ExcThread",
    "ThreadTaskDoneCallback",
    "ThreadDoneCallback",
    "TaskDoneCallback",
    "current_task_or_thread",
    "to_thread",
    "agen_with_wait",
    "merge_aiters",
    "profile_func",
    "run_in_process",
    "RunInProcess",
]

from .pubsub import PubSub, PubSubItem
from .thread_task_id import ThreadTaskIdComposer
from .thread_safe_event import ThreadSafeAsyncioEvent
from .loop import ToLoop
from .run import RunInProcess, run_in_process
from .multiprocessing_logging import MultiprocessingLogging
from .peek import peek_stdout, peek_stderr, peek_textio
from .thread_exception import ExcThread
from .done_callback import (
    ThreadTaskDoneCallback,
    ThreadDoneCallback,
    TaskDoneCallback,
)
from .func import (
    current_task_or_thread,
    to_thread,
    agen_with_wait,
    merge_aiters,
)
from .profile import profile_func



import time
import random
import asyncio
import threading

from nextline.utils import TaskDoneCallback

import pytest
from unittest.mock import Mock


async def target(obj: TaskDoneCallback):
    assert asyncio.current_task() == obj.register()
    delay = random.random() * 0.01
    await asyncio.sleep(delay)


class Done:
    """A callback function"""

    def __init__(self):
        self.args = set()

    def __call__(self, arg):
        self.args.add(arg)


@pytest.fixture()
def done():
    """A callback function"""
    yield Done()


async def test_aclose(done: Done):
    obj = TaskDoneCallback(done=done)
    t = asyncio.create_task(target(obj))
    await t
    await obj.aclose()
    assert {t} == done.args


async def test_async_with(done: Done):
    async with TaskDoneCallback(done=done) as obj:
        t = asyncio.create_task(target(obj))
        await t
    assert {t} == done.args


def test_asyncio_run_close(done: Done):
    obj = TaskDoneCallback(done=done)
    asyncio.run(target(obj))
    obj.close()
    assert 1 == len(done.args)


def test_asyncio_run_with(done: Done):
    with TaskDoneCallback(done=done) as obj:
        asyncio.run(target(obj))
    assert 1 == len(done.args)


def test_thread(done: Done):
    event = threading.Event()

    def f(obj: TaskDoneCallback):
        asyncio.run(target(obj))
        event.set()

    with TaskDoneCallback(done=done) as obj:
        t = threading.Thread(target=f, args=(obj,))
        t.start()
        event.wait()

    assert 1 == len(done.args)
    t.join()


async def test_register_arg(done: Done):
    async def target():
        delay = random.random() * 0.01
        await asyncio.sleep(delay)

    async with TaskDoneCallback(done=done) as obj:
        t = asyncio.create_task(target())

        # manually provide the task object
        assert t == obj.register(t)

        await t

    assert {t} == done.args


@pytest.mark.parametrize("n_tasks", [0, 1, 2, 5, 10])
async def test_multiple(n_tasks: int, done: Done):

    async with TaskDoneCallback(done=done) as obj:
        tasks = {asyncio.create_task(target(obj)) for _ in range(n_tasks)}
        await asyncio.gather(*tasks)

    assert tasks == done.args


async def test_raise_aclose_from_task(done: Done):
    async def target(obj: TaskDoneCallback):
        obj.register()
        delay = random.random() * 0.01
        time.sleep(delay)
        await obj.aclose()

    obj = TaskDoneCallback(done=done)
    t = asyncio.create_task(target(obj))

    with pytest.raises(RuntimeError):
        await t


async def test_raise_close_from_task(done: Done):
    async def target(obj: TaskDoneCallback):
        obj.register()
        delay = random.random() * 0.01
        time.sleep(delay)
        obj.close()

    obj = TaskDoneCallback(done=done)
    t = asyncio.create_task(target(obj))

    with pytest.raises(RuntimeError):
        await t


async def test_raise_in_done():
    done = Mock(side_effect=ValueError)
    obj = TaskDoneCallback(done=done)
    t = asyncio.create_task(target(obj))

    await t

    with pytest.raises(ValueError):
        await obj.aclose()


async def test_done_none():
    async with TaskDoneCallback() as obj:
        t = asyncio.create_task(target(obj))
        await asyncio.sleep(0)  # let the task be registered
    assert t.done()  # finished after exited
    await t



from __future__ import annotations

from concurrent.futures import ProcessPoolExecutor
import logging
from pytest import LogCaptureFixture
from nextline.utils import MultiprocessingLogging


async def test_multiprocessing_logging(caplog: LogCaptureFixture):
    with caplog.at_level(logging.DEBUG):
        async with MultiprocessingLogging() as mp_logging:
            with ProcessPoolExecutor(
                1, initializer=mp_logging.init
            ) as executor:
                fut = executor.submit(fn)
                assert "foo" == fut.result()

    assert len(caplog.records) == 1
    assert caplog.records[0].levelname == "DEBUG"
    assert caplog.records[0].message == "bar"
    assert caplog.records[0].name == __name__


def fn():
    logger = logging.getLogger(__name__)
    logger.debug("bar")
    return "foo"
